# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test the component that manages drop-in configuration.
"""

from unittest.mock import Mock, patch

import pytest

from plinth.app import App
from plinth.config import DropinConfigs
from plinth.diagnostic_check import DiagnosticCheck, Result

pytestmark = pytest.mark.usefixtures('mock_privileged')
privileged_modules_to_mock = ['plinth.privileged.config']


@pytest.fixture(name='dropin_configs')
def fixture_dropin_configs():
    """Fixture to create a basic drop-in configs component."""

    class AppTest(App):
        app_id = 'test-app'

    app = AppTest()
    component = DropinConfigs('test-component',
                              ['/etc/test/path1', '/etc/path2'])
    app.add(component)

    return component


@pytest.fixture(autouse=True)
def fixture_assert_dropin_config():
    """Mock asserting dropin config path."""
    with patch('plinth.privileged.config._assert_managed_dropin_config'):
        yield


def test_dropin_configs_init(dropin_configs):
    """Test initialization of drop-in configs component."""
    assert dropin_configs.component_id == 'test-component'
    assert dropin_configs.etc_paths[0] == '/etc/test/path1'
    assert dropin_configs.etc_paths[1] == '/etc/path2'
    assert not dropin_configs.copy_only

    component = DropinConfigs('test-component', [], copy_only=False)
    assert not component.copy_only

    component = DropinConfigs('test-component', [], copy_only=True)
    assert component.copy_only


def _assert_symlinks(component, tmp_path, should_exist, copy_only=False):
    """Assert that symlinks exists and they point correctly."""
    for path in component.etc_paths:
        full_path = tmp_path / path.lstrip('/')
        if should_exist:
            target = tmp_path / 'usr/share/freedombox' / path.lstrip('/')
            if copy_only:
                assert full_path.is_file()
                assert full_path.read_text() == target.read_text()
            else:
                assert full_path.is_symlink()
                assert full_path.resolve() == target
        else:
            assert not full_path.exists()


def test_dropin_configs_setup(dropin_configs, tmp_path):
    """Test setup for dropin configs component."""
    with patch('plinth.config.DropinConfigs.ROOT', new=tmp_path):
        is_enabled = Mock()
        App.get('test-app').is_enabled = is_enabled

        is_enabled.return_value = False
        dropin_configs.setup(old_version=0)
        _assert_symlinks(dropin_configs, tmp_path, should_exist=False)

        is_enabled.return_value = True
        dropin_configs.setup(old_version=0)
        _assert_symlinks(dropin_configs, tmp_path, should_exist=True)


def test_dropin_configs_enable_disable_symlinks(dropin_configs, tmp_path):
    """Test enable/disable for dropin configs component for symlinks."""
    with patch('plinth.config.DropinConfigs.ROOT', new=tmp_path):
        # Enable when nothing exists
        dropin_configs.enable()
        _assert_symlinks(dropin_configs, tmp_path, should_exist=True)

        # Disable
        dropin_configs.disable()
        _assert_symlinks(dropin_configs, tmp_path, should_exist=False)

        # Enable when a file already exists
        dropin_configs.disable()
        etc_path = DropinConfigs._get_etc_path('/etc/test/path1')
        etc_path.touch()
        dropin_configs.enable()
        _assert_symlinks(dropin_configs, tmp_path, should_exist=True)

        # When symlink already exists to wrong location
        dropin_configs.disable()
        etc_path.symlink_to('/blah')
        dropin_configs.enable()
        _assert_symlinks(dropin_configs, tmp_path, should_exist=True)

        # When symlink already exists to correct location
        dropin_configs.disable()
        target_path = DropinConfigs._get_target_path('/etc/test/path1')
        etc_path.symlink_to(target_path)
        dropin_configs.enable()
        _assert_symlinks(dropin_configs, tmp_path, should_exist=True)


def test_dropin_configs_enable_disable_copy_only(dropin_configs, tmp_path):
    """Test enable/disable for dropin configs component for copying."""
    with patch('plinth.config.DropinConfigs.ROOT', new=tmp_path):
        dropin_configs.copy_only = True
        for path in ['/etc/test/path1', '/etc/path2']:
            target = DropinConfigs._get_target_path(path)
            target.parent.mkdir(parents=True, exist_ok=True)
            target.write_text('test-config-content')

        # Enable when nothing exists
        dropin_configs.enable()
        _assert_symlinks(dropin_configs, tmp_path, should_exist=True,
                         copy_only=True)

        # Disable
        dropin_configs.disable()
        _assert_symlinks(dropin_configs, tmp_path, should_exist=False,
                         copy_only=True)

        # Enable when a file already exists with wrong content
        dropin_configs.disable()
        etc_path = DropinConfigs._get_etc_path('/etc/test/path1')
        etc_path.write_text('x-invalid-content')
        dropin_configs.enable()
        _assert_symlinks(dropin_configs, tmp_path, should_exist=True,
                         copy_only=True)

        # When the file is a symlink
        dropin_configs.disable()
        etc_path.symlink_to('/blah')
        dropin_configs.enable()
        _assert_symlinks(dropin_configs, tmp_path, should_exist=True,
                         copy_only=True)

        # When copy already exists with correct content
        dropin_configs.disable()
        etc_path.write_text('test-config-content')
        dropin_configs.enable()
        _assert_symlinks(dropin_configs, tmp_path, should_exist=True,
                         copy_only=True)


def test_dropin_config_diagnose_symlinks(dropin_configs, tmp_path):
    """Test diagnosing dropin configs for symlinks."""
    with patch('plinth.config.DropinConfigs.ROOT', new=tmp_path):
        # Nothing exists
        results = dropin_configs.diagnose()
        assert results == [
            DiagnosticCheck(
                f'dropin-config-{tmp_path}/etc/test/path1',
                'Static configuration {etc_path} is setup properly',
                Result.FAILED, {'etc_path': f'{tmp_path}/etc/test/path1'}),
            DiagnosticCheck(
                f'dropin-config-{tmp_path}/etc/path2',
                'Static configuration {etc_path} is setup properly',
                Result.FAILED, {'etc_path': f'{tmp_path}/etc/path2'}),
        ]

        # Proper symlinks exist
        dropin_configs.enable()
        results = dropin_configs.diagnose()
        assert results[0].result == 'passed'
        assert results[1].result == 'passed'

        # A file exists instead of symlink
        dropin_configs.disable()
        etc_path = DropinConfigs._get_etc_path('/etc/test/path1')
        etc_path.touch()
        results = dropin_configs.diagnose()
        assert results[0].result == 'failed'

        # Symlink points to wrong location
        dropin_configs.disable()
        etc_path.symlink_to('/blah')
        results = dropin_configs.diagnose()
        assert results[0].result == 'failed'

        # Symlink is recreated
        dropin_configs.enable()
        results = dropin_configs.diagnose()
        assert results[0].result == 'passed'


def test_dropin_config_diagnose_copy_only(dropin_configs, tmp_path):
    """Test diagnosing dropin configs."""
    with patch('plinth.config.DropinConfigs.ROOT', new=tmp_path):
        dropin_configs.copy_only = True
        for path in ['/etc/test/path1', '/etc/path2']:
            target = DropinConfigs._get_target_path(path)
            target.parent.mkdir(parents=True, exist_ok=True)
            target.write_text('test-config-content')

        # Nothing exists
        results = dropin_configs.diagnose()
        assert results[0].result == 'failed'
        assert results[1].result == 'failed'

        # Proper copies exist
        dropin_configs.enable()
        results = dropin_configs.diagnose()
        assert results[0].result == 'passed'
        assert results[1].result == 'passed'

        # A symlink exists instead of a copied file
        dropin_configs.disable()
        etc_path = DropinConfigs._get_etc_path('/etc/test/path1')
        etc_path.symlink_to('/blah')
        results = dropin_configs.diagnose()
        assert results[0].result == 'failed'

        # Copied file contains wrong contents
        dropin_configs.disable()
        etc_path.write_text('x-invalid-contents')
        results = dropin_configs.diagnose()
        assert results[0].result == 'failed'



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Views for WireGuard application.
"""

import urllib.parse

from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.http import Http404
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.translation import gettext as _
from django.views.generic import FormView, TemplateView

from plinth import network
from plinth.modules.names.components import DomainName
from plinth.views import AppView

from . import forms, utils


class WireguardView(AppView):
    """Serve configuration page."""
    app_id = 'wireguard'
    diagnostics_module_name = 'wireguard'
    template_name = 'wireguard.html'

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        info = utils.get_info()
        context['server'] = info['my_server']
        context['client_peers'] = info['my_client']['servers']
        return context


class AddClientView(SuccessMessageMixin, FormView):
    """View to add a client."""
    form_class = forms.AddClientForm
    template_name = 'wireguard_add_client.html'
    success_url = reverse_lazy('wireguard:index')
    success_message = _('Added new client.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Add Allowed Client')
        return context

    def form_valid(self, form):
        """Add the client."""
        public_key = form.cleaned_data.get('public_key')
        try:
            utils.add_client(public_key)
        except ValueError:
            messages.warning(self.request,
                             _('Client with public key already exists'))
            return redirect('wireguard:index')

        return super().form_valid(form)


class ShowClientView(SuccessMessageMixin, TemplateView):
    """View to show a client's details."""
    template_name = 'wireguard_show_client.html'

    def get_context_data(self, **kwargs):
        """Return additional context data for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Allowed Client')

        public_key = urllib.parse.unquote(self.kwargs['public_key'])
        server_info = utils.get_info()['my_server']
        if not server_info or public_key not in server_info['peers']:
            raise Http404

        domains = DomainName.list_names(filter_for_service='wireguard')
        context['server'] = server_info
        context['client'] = server_info['peers'][public_key]
        context['endpoints'] = [
            domain + ':' + str(server_info['listen_port'])
            for domain in domains
        ]
        return context


class EditClientView(SuccessMessageMixin, FormView):
    """View to modify a client."""
    form_class = forms.AddClientForm
    template_name = 'wireguard_edit_client.html'
    success_url = reverse_lazy('wireguard:index')
    success_message = _('Updated client.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Modify Client')
        return context

    def get_initial(self):
        """Get initial form data."""
        initial = super().get_initial()
        initial['public_key'] = urllib.parse.unquote(self.kwargs['public_key'])
        return initial

    def form_valid(self, form):
        """Update the client."""
        old_public_key = form.initial['public_key']
        public_key = form.cleaned_data.get('public_key')

        if old_public_key != public_key:
            try:
                utils.add_client(public_key)
            except ValueError:
                messages.warning(self.request,
                                 _('Client with public key already exists'))

            utils.remove_client(old_public_key)

        return super().form_valid(form)


class DeleteClientView(SuccessMessageMixin, TemplateView):
    """View to delete a client."""
    template_name = 'wireguard_delete_client.html'

    def get_context_data(self, **kwargs):
        """Return additional context data for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Delete Allowed Client')
        context['public_key'] = urllib.parse.unquote(self.kwargs['public_key'])
        return context

    def post(self, request, public_key):
        """Delete the client."""
        public_key = urllib.parse.unquote(public_key)
        try:
            utils.remove_client(public_key)
            messages.success(request, _('Client deleted.'))
        except KeyError:
            messages.error(request, _('Client not found'))

        return redirect('wireguard:index')


class AddServerView(SuccessMessageMixin, FormView):
    """View to add a server."""
    form_class = forms.AddServerForm
    template_name = 'wireguard_add_server.html'
    success_url = reverse_lazy('wireguard:index')
    success_message = _('Added new server.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Add Connection to Server')
        return context

    def form_valid(self, form):
        """Add the server."""
        utils.add_server(form.get_settings())
        return super().form_valid(form)


class ShowServerView(SuccessMessageMixin, TemplateView):
    """View to show a server's details."""
    template_name = 'wireguard_show_server.html'

    def get_context_data(self, **kwargs):
        """Return additional context data for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Connection to Server')

        interface = self.kwargs['interface']
        info = utils.get_info()
        server = info['my_client']['servers'].get(interface)
        if not server:
            raise Http404

        context['interface'] = interface
        context['server'] = server
        return context


class EditServerView(SuccessMessageMixin, FormView):
    """View to modify a server."""
    form_class = forms.AddServerForm
    template_name = 'wireguard_edit_server.html'
    success_url = reverse_lazy('wireguard:index')
    success_message = _('Updated server.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Modify Connection to Server')
        return context

    def get_initial(self):
        """Get initial form data."""
        initial = super().get_initial()
        interface = self.kwargs['interface']
        info = utils.get_nm_info()
        server = info.get(interface)
        if not server:
            raise Http404

        initial['ip_address'] = server.get('ip_address')
        if server['peers']:
            peer = next(peer for peer in server['peers'].values())
            initial['peer_endpoint'] = peer['endpoint']
            initial['peer_public_key'] = peer['public_key']
            initial['private_key'] = server['private_key']
            initial['preshared_key'] = peer['preshared_key']
            initial['default_route'] = server['default_route']

        return initial

    def form_valid(self, form):
        """Update the server."""
        interface = self.kwargs['interface']
        utils.edit_server(interface, form.get_settings())
        return super().form_valid(form)


class DeleteServerView(SuccessMessageMixin, TemplateView):
    """View to delete a server."""
    template_name = 'wireguard_delete_server.html'

    def get_context_data(self, **kwargs):
        """Return additional context data for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Delete Connection to Server')

        interface = self.kwargs['interface']
        info = utils.get_nm_info()
        server = info.get(interface)
        if not server:
            raise Http404

        context['interface'] = interface
        if server['peers']:
            peer = next(peer for peer in server['peers'].values())
            context['peer_endpoint'] = peer['endpoint']
            context['peer_public_key'] = peer['public_key']

        return context

    def post(self, request, interface):
        """Delete the server."""
        connection = network.get_connection_by_interface_name(interface)
        network.delete_connection(connection.get_uuid())
        messages.success(request, _('Server deleted.'))
        return redirect('wireguard:index')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test actions for configuring bind
"""
from pathlib import Path

import pytest

from plinth.modules import bind


@pytest.fixture(name='configuration_file')
def fixture_configuration_file(tmp_path):
    """Setup the a bind configuration file temporary directory."""
    conf_file = tmp_path / 'named.conf.options'
    conf_file.write_text(bind.privileged.DEFAULT_CONFIG)
    old_config_file = bind.privileged.CONFIG_FILE
    bind.privileged.CONFIG_FILE = str(conf_file)
    yield
    bind.privileged.CONFIG_FILE = old_config_file


@pytest.fixture
def bind_zones_folder(tmp_path):
    """Setup the a bind configuration file temporary directory."""
    test_zone_file = """
;
; BIND data file for local loopback interface
;
$TTL	604800
@	IN	SOA	{name} root.localhost. (
			      2		; Serial
			 604800		; Refresh
			  86400		; Retry
			2419200		; Expire
			 604800 )	; Negative Cache TTL
;
@	IN	NS	{name}
@	IN	A	{a_record}
@	IN	AAAA	{aaaa_record}
    """ # noqa

    old_zones_dir = bind.privileged.ZONES_DIR
    bind.privileged.ZONES_DIR = tmp_path
    zones_dir_path = Path(bind.privileged.ZONES_DIR)
    zones_dir_path.mkdir(exist_ok=True, parents=True)

    local_path = zones_dir_path / "local.zone"
    local_path.write_text(
        test_zone_file.format(name='localhost.', a_record="127.0.0.1",
                              aaaa_record="::1"))

    custom_zone_path = zones_dir_path / "custom.zone"
    custom_zone_path.write_text(
        test_zone_file.format(name='custom.domain.', a_record="10.10.10.1",
                              aaaa_record="fe80::c6e9:84ff:fe16:95da"))

    yield

    local_path.unlink()
    custom_zone_path.unlink()
    bind.privileged.ZONES_DIR = old_zones_dir


@pytest.mark.usefixtures('configuration_file')
def test_set_forwarders():
    """Test that setting forwarders works."""
    bind.privileged._set_forwarders('8.8.8.8 8.8.4.4')
    conf = bind.privileged.get_config()
    assert conf['forwarders'] == '8.8.8.8 8.8.4.4'

    bind.privileged._set_forwarders('')
    conf = bind.privileged.get_config()
    assert conf['forwarders'] == ''


@pytest.mark.usefixtures('bind_zones_folder')
def test_get_correct_served_domains():
    """
    Test that get_served_domains collects the right a/aaaa records from zone
    files
    """
    served_domains = bind.privileged.get_served_domains()

    assert served_domains['localhost.'] == ["127.0.0.1", "::1"]
    assert served_domains['custom.domain.'] == [
        "10.10.10.1", "fe80::c6e9:84ff:fe16:95da"
    ]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for Gitweb."""

import configparser
import logging
import os
import pathlib
import re
import shutil
import subprocess
import time
from typing import Any

from plinth import action_utils
from plinth.actions import privileged
from plinth.modules.gitweb.forms import RepositoryValidator, get_name_from_url
from plinth.modules.gitweb.manifest import GIT_REPO_PATH, REPO_DIR_OWNER

logger = logging.getLogger(__name__)


def validate_repo_name(name: str) -> str:
    """Validate a repository name and add .git extension if necessary."""
    RepositoryValidator()(name)
    if not name.endswith('.git'):
        name = name + '.git'

    return name


def validate_repo_url(url: str) -> str:
    """Validate a repository URL."""
    RepositoryValidator(input_should_be='url')(url)
    return url


@privileged
def setup():
    """Disable default Apache2 Gitweb configuration."""
    action_utils.webserver_disable('gitweb')
    if not _get_global_default_branch():
        _set_global_default_branch('main')


def _get_global_default_branch():
    """Get globally configured default branch name."""
    try:
        default_branch = subprocess.check_output(
            ['git', 'config', '--global', '--get',
             'init.defaultBranch']).decode().strip()
    except subprocess.CalledProcessError as exception:
        if exception.returncode == 1:  # Default branch not configured
            return None
        raise

    return default_branch


def _set_global_default_branch(name):
    """Configure default branch name globally."""
    subprocess.check_call(
        ['git', 'config', '--global', 'init.defaultBranch', name])


def _clone_with_progress_report(url, repo_dir):
    """Clone a repository and write progress info to the file."""
    starttime = time.time()
    status_file = os.path.join(repo_dir, 'clone_progress')
    repo_temp_dir = os.path.join(repo_dir, '.temp')
    # do not ask for credidentials and set low speed timeout
    env = dict(os.environ, GIT_TERMINAL_PROMPT='0', LC_ALL='C',
               GIT_HTTP_LOW_SPEED_LIMIT='100', GIT_HTTP_LOW_SPEED_TIME='60')

    proc = subprocess.Popen(
        ['git', 'clone', '--bare', '--progress', url, repo_temp_dir],
        stderr=subprocess.PIPE, text=True, env=env)

    # write clone progress to the file
    errors = []
    while True:
        line = proc.stderr.readline()
        if not line:
            break

        if 'error:' in line or 'fatal:' in line:
            errors.append(line.strip())

        currenttime = time.time()
        if currenttime - starttime > 1:
            elapsed = _clone_status_line_to_percent(line)
            if elapsed is not None:
                try:
                    with open(status_file, 'w',
                              encoding='utf-8') as file_handle:
                        file_handle.write(elapsed)
                except OSError as error:
                    errors.append(str(error))

            starttime = currenttime

    # make sure process is ended
    try:
        proc.communicate(timeout=10)
    except subprocess.TimeoutExpired:
        proc.kill()

    os.remove(status_file)
    if proc.returncode != 0:
        shutil.rmtree(repo_dir)
        logger.error('Git repository cloning failed: %s', '\n'.join(errors))
        raise RuntimeError('Git repository cloning failed.', errors)


def _prepare_clone_repo(url: str, is_private: bool):
    """Prepare cloning a repository."""
    repo_name = get_name_from_url(url)
    if not repo_name.endswith('.git'):
        repo_name = repo_name + '.git'

    repo_dir = os.path.join(GIT_REPO_PATH, repo_name)
    os.mkdir(repo_dir)

    status_file = os.path.join(repo_dir, 'clone_progress')
    try:
        if is_private:
            _set_access_status(repo_name, 'private')
        with open(status_file, 'w', encoding='utf-8') as file_handle:
            file_handle.write('0')
    except OSError:
        shutil.rmtree(repo_dir)
        raise


def _clone_status_line_to_percent(line):
    """Parse Git clone command output."""
    result = re.match(r'.* ([0-9]+)% ', line)
    if result is not None:
        text = result.group(0)
        progress = int(result.group(1))
        if 'Counting objects' in text:
            total_progress = 0.05 * progress
        elif 'Compressing objects' in text:
            total_progress = 5 + 0.05 * progress
        elif 'Receiving objects' in text:
            total_progress = 10 + 0.6 * progress
        elif 'Resolving deltas' in text:
            total_progress = 70 + 0.3 * progress
        else:
            return None

        return str(int(total_progress))

    return None


def _clone_repo(url: str, description: str, owner: str, keep_ownership: bool):
    """Clone a repository."""
    repo = get_name_from_url(url)
    if not repo.endswith('.git'):
        repo = repo + '.git'

    repo_path = os.path.join(GIT_REPO_PATH, repo)
    repo_temp_path = os.path.join(repo_path, '.temp')

    _clone_with_progress_report(url, repo_path)

    for item in os.listdir(repo_temp_path):
        shutil.move(os.path.join(repo_temp_path, item), repo_path)

    shutil.rmtree(repo_temp_path)
    if not keep_ownership:
        subprocess.check_call(
            ['chown', '-R', f'{REPO_DIR_OWNER}:{REPO_DIR_OWNER}', repo],
            cwd=GIT_REPO_PATH)

    _set_repo_description(repo, description)
    _set_repo_owner(repo, owner)


def _create_repo(repo: str, description: str, owner: str, is_private: bool,
                 keep_ownership: bool):
    """Create an empty repository."""
    try:
        subprocess.check_call(['git', 'init', '-q', '--bare', repo],
                              cwd=GIT_REPO_PATH)
        if not keep_ownership:
            subprocess.check_call(
                ['chown', '-R', f'{REPO_DIR_OWNER}:{REPO_DIR_OWNER}', repo],
                cwd=GIT_REPO_PATH)
        _set_repo_description(repo, description)
        _set_repo_owner(repo, owner)
        if is_private:
            _set_access_status(repo, 'private')
    except (subprocess.CalledProcessError, OSError):
        repo_path = os.path.join(GIT_REPO_PATH, repo)
        if os.path.isdir(repo_path):
            shutil.rmtree(repo_path)
        raise


def _get_default_branch(repo):
    """Get default branch of the repository."""
    repo_path = os.path.join(GIT_REPO_PATH, repo)

    return subprocess.check_output(
        ['git', '-C', repo_path, 'symbolic-ref', '--short',
         'HEAD']).decode().strip()


def _get_repo_description(repo):
    """Set description of the repository."""
    description_file = os.path.join(GIT_REPO_PATH, repo, 'description')
    if os.path.exists(description_file):
        with open(description_file, 'r', encoding='utf-8') as file_handle:
            description = file_handle.read()
    else:
        description = ''

    return description


def _set_repo_description(repo, description):
    """Set description of the repository."""
    description_file = os.path.join(GIT_REPO_PATH, repo, 'description')
    with open(description_file, 'w', encoding='utf-8') as file_handle:
        file_handle.write(description)


def _get_repo_owner(repo):
    """Set repository's owner name."""
    repo_config = os.path.join(GIT_REPO_PATH, repo, 'config')
    config = configparser.ConfigParser()
    config.read(repo_config)
    try:
        owner = config['gitweb']['owner']
    except KeyError:
        owner = ''

    return owner


def _set_repo_owner(repo, owner):
    """Set repository's owner name."""
    repo_config = os.path.join(GIT_REPO_PATH, repo, 'config')
    config = configparser.ConfigParser()
    config.read(repo_config)
    if not config.has_section('gitweb'):
        config.add_section('gitweb')

    config['gitweb']['owner'] = owner
    with open(repo_config, 'w', encoding='utf-8') as file_handle:
        config.write(file_handle)


def _get_access_status(repo):
    """Get repository's access status."""
    private_file = os.path.join(GIT_REPO_PATH, repo, 'private')
    if os.path.exists(private_file):
        return 'private'

    return 'public'


def _set_access_status(repo, status):
    """Set repository as private or public."""
    private_file = os.path.join(GIT_REPO_PATH, repo, 'private')
    if status == 'private':
        open(private_file, 'a', encoding='utf-8')
    elif status == 'public':
        if os.path.exists(private_file):
            os.remove(private_file)


def _get_branches(repo):
    """Return list of the branches in the repository."""
    output = subprocess.check_output(
        ['git', '-C', repo, 'branch', '--format=%(refname:short)'],
        cwd=GIT_REPO_PATH)

    return output.decode().strip().split()


@privileged
def get_branches(name: str) -> dict[str, Any]:
    """Check whether a branch exists in the repository."""
    repo = validate_repo_name(name)
    return dict(default_branch=_get_default_branch(repo),
                branches=_get_branches(repo))


@privileged
def rename_repo(old_name: str, new_name: str):
    """Rename a repository."""
    old_name = validate_repo_name(old_name)
    new_name = validate_repo_name(new_name)
    oldpath = os.path.join(GIT_REPO_PATH, old_name)
    newpath = os.path.join(GIT_REPO_PATH, new_name)
    os.rename(oldpath, newpath)


@privileged
def set_default_branch(name: str, branch: str):
    """Set description of the repository."""
    repo = validate_repo_name(name)
    if branch not in _get_branches(repo):
        raise ValueError('No such branch')

    subprocess.check_call([
        'git', '-C', repo, 'symbolic-ref', 'HEAD',
        "refs/heads/{}".format(branch)
    ], cwd=GIT_REPO_PATH)


@privileged
def set_repo_description(name: str, description: str):
    """Set description of the repository."""
    repo = validate_repo_name(name)
    _set_repo_description(repo, description)


@privileged
def set_repo_owner(name: str, owner: str):
    """Set repository's owner name."""
    repo = validate_repo_name(name)
    _set_repo_owner(repo, owner)


@privileged
def set_repo_access(name: str, access: str):
    """Set repository's access status."""
    repo = validate_repo_name(name)
    if access not in ('public', 'private'):
        raise ValueError('Invalid access parameter')

    _set_access_status(repo, access)


@privileged
def repo_info(name: str) -> dict[str, str]:
    """Get information about repository."""
    repo = validate_repo_name(name)
    repo_path = os.path.join(GIT_REPO_PATH, repo)
    if not os.path.exists(repo_path):
        raise RuntimeError('Repository not found')

    return dict(name=repo[:-4], description=_get_repo_description(repo),
                owner=_get_repo_owner(repo), access=_get_access_status(repo),
                default_branch=_get_default_branch(repo))


@privileged
def create_repo(url: str | None = None, name: str | None = None,
                description: str = '', owner: str = '',
                keep_ownership: bool = False, is_private: bool = False,
                skip_prepare: bool = False, prepare_only: bool = False):
    """Create a new or clone a remote repository."""
    if url:
        url = validate_repo_url(url)

    if name:
        repo = validate_repo_name(name)

    if url:
        if not skip_prepare:
            _prepare_clone_repo(url, is_private)

        if not prepare_only:
            _clone_repo(url, description, owner, keep_ownership)
    elif repo is not None:
        _create_repo(repo, description, owner, is_private, keep_ownership)


@privileged
def repo_exists(url: str) -> bool:
    """Return whether remote repository exists."""
    url = validate_repo_url(url)
    env = dict(os.environ, GIT_TERMINAL_PROMPT='0')
    try:
        subprocess.check_call(['git', 'ls-remote', url, 'HEAD'], timeout=10,
                              env=env)
        return True
    except subprocess.CalledProcessError:
        return False


@privileged
def delete_repo(name: str):
    """Delete a git repository."""
    repo = validate_repo_name(name)
    repo_path = os.path.join(GIT_REPO_PATH, repo)
    shutil.rmtree(repo_path)


@privileged
def uninstall():
    """Remove git repositories."""
    for item in pathlib.Path(GIT_REPO_PATH).glob('*'):
        shutil.rmtree(item, ignore_errors=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for zoph app.
"""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.zoph]


class TestZophApp(functional.BaseAppTests):
    app_name = 'zoph'
    has_service = False
    has_web = True

    def install_and_setup(self, session_browser):
        """Install the app and run setup."""
        super().install_and_setup(session_browser)
        self._zoph_is_setup(session_browser)

    def _zoph_is_setup(self, session_browser):
        """Click setup button on the setup page."""
        functional.nav_to_module(session_browser, self.app_name)
        if session_browser.find_by_css('.form-configuration'):
            functional.submit(session_browser, form_class='form-configuration')

    def assert_app_running(self, session_browser):
        assert functional.app_is_enabled(session_browser, self.app_name)

    def assert_app_not_running(self, session_browser):
        assert not functional.app_is_enabled(session_browser, self.app_name)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for transmission app.
"""

import os

import pytest
from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.transmission, pytest.mark.sso]


class TestTransmissionApp(functional.BaseAppTests):
    app_name = 'transmission'
    has_service = False
    has_web = True

    def test_upload_torrent(self, session_browser):
        """Test uploading a torrent to Transmission."""
        functional.app_enable(session_browser, 'transmission')
        _remove_all_torrents(session_browser)
        _upload_sample_torrent(session_browser)
        _assert_number_of_torrents(session_browser, 1)

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test backup and restore of app data."""
        functional.app_enable(session_browser, 'transmission')
        _remove_all_torrents(session_browser)
        _upload_sample_torrent(session_browser)
        functional.backup_create(session_browser, 'transmission',
                                 'test_transmission')

        _remove_all_torrents(session_browser)
        functional.backup_restore(session_browser, 'transmission',
                                  'test_transmission')

        assert functional.service_is_running(session_browser, 'transmission')
        _assert_number_of_torrents(session_browser, 1)


def _assert_number_of_torrents(session_browser, torrents_number):
    functional.visit(session_browser, '/transmission')
    assert functional.eventually(
        lambda: torrents_number == _get_number_of_torrents(session_browser))


def _remove_all_torrents(browser):
    """Remove all torrents from transmission."""
    functional.visit(browser, '/transmission')
    while True:
        torrents = browser.find_by_css('#torrent_list .torrent')
        if not torrents:
            break

        torrents.first.click()
        functional.eventually(browser.is_element_not_present_by_css,
                              args=['#toolbar-remove.disabled'])
        browser.click_link_by_id('toolbar-remove')
        functional.eventually(
            browser.is_element_not_present_by_css,
            args=['#dialog-container[style="display: none;"]'])
        browser.click_link_by_id('dialog_confirm_button')
        functional.eventually(browser.is_element_present_by_css,
                              args=['#toolbar-remove.disabled'])


def _upload_sample_torrent(browser):
    """Upload a sample torrent into transmission."""
    functional.visit(browser, '/transmission')
    file_path = os.path.join(os.path.dirname(__file__), 'data',
                             'sample.torrent')
    browser.click_link_by_id('toolbar-open')
    functional.eventually(browser.is_element_not_present_by_css,
                          args=['#upload-container[style="display: none;"]'])
    browser.attach_file('torrent_files[]', [file_path])
    browser.click_link_by_id('upload_confirm_button')
    functional.eventually(browser.is_element_present_by_css,
                          args=['#torrent_list .torrent'])


def _get_number_of_torrents(browser):
    """Return the number torrents currently in transmission."""
    return len(browser.find_by_css('#torrent_list .torrent'))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Common utilities to help with handling a database.
"""

import threading
from typing import ClassVar


class DBLock:
    """A re-entrant lock with a fixed timeout (not -1) by default."""

    TIMEOUT: ClassVar[float] = 30

    def __init__(self, *args, **kwargs):
        """Create an RLock object."""
        self._lock = threading.RLock(*args, **kwargs)
        self.timeout = DBLock.TIMEOUT

    def __getattr__(self, name):
        """Return RLock attributes."""
        return getattr(self._lock, name)

    def __enter__(self):
        """Use RLock context management."""
        return self._lock.acquire(timeout=self.timeout)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Use RLock context management."""
        try:
            self._lock.release()
        except RuntimeError:
            # Lock was not acquired
            pass


# The Problem
# -----------
# Since this is a small application (in terms of data stored and requests per
# second handled), we use sqlite3 as the backend database. This gives us more
# reliability. However, sqlite3 has limited capability to handle parallel
# requests. When we try to use multiple Model.update_or_create() calls
# simultaneously, only the first one succeeds and the remaining fail with
# 'Database is locked' errors.
#
# The 'timeout' value passed during connection creation means that queries will
# wait for a maximum of given timeout period when trying to acquire locks.
# However, in some cases, to prevent deadlocks caused by threads waiting on
# each other, sqlite3 will not wait and immediately throw an exception.
#
# If we set the isolation level to 'EXCLUSIVE' or 'IMMEDIATE' instead of the
# default 'DEFERRED', each transaction should acquire a write lock immediately
# after transaction is started. This reduces the situations in which threads
# deadlock on each other. This improves the situation somewhat. However, this
# is still causing immediate timeout errors due to the way Django does its
# update_or_create() queries.
#
# This is true even in 'WAL' (write-ahead-log) journaling mode. While WAL mode
# makes it easier to have many read transactions while one write query is
# happening, it does not prevent this situation.
#
# Django should ideally provide a way to serialize these queries specifically
# to work with sqlite3 locking behavior. There seems to be no such way
# currently.
#
# Workaround
# ----------
# To workaround the problem, use a simple lock to serialize all database
# queries. Like this:
#
# with db.lock:
#     # do some short database operation
#
# Queries usually have short execution time (unless stuck on disk I/O). A lot
# of parallel request processing is not expected from this service. We want
# more reliability (not failing on DB locks) over the ability to run parallel
# requests. Typically, requests waiting to perform DB queries will only have to
# wait < 1 second to get their turn. In the worst case, the database lock will
# not be acquired and the code continues to anyway.
#
# This locking can't be done in all situations. For example, queries made
# within Django framework are not locked. Still, this approach should prevent
# most of the significant cases where we have seen database lock issues.

lock = DBLock()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for BitTorrent web client."""

import pathlib
import shutil
import subprocess
import time

import augeas

from plinth import action_utils
from plinth.actions import privileged
from plinth.modules.deluge.utils import Config

DELUGED_DEFAULT_FILE = '/etc/default/deluged'
DELUGE_CONF_DIR = pathlib.Path('/var/lib/deluged/.config/deluge/')

DELUGE_WEB_SYSTEMD_SERVICE_PATH = '/etc/systemd/system/deluge-web.service'
DELUGE_WEB_SYSTEMD_SERVICE = '''
#
# This file is managed and overwritten by Plinth.  If you wish to edit
# it, disable Deluge in Plinth, remove this file and manage it manually.
#
[Unit]
Description=Deluge Web Interface
Documentation=man:deluge-web(1)
After=network.target

[Service]
ExecStart=bash -c "/usr/bin/deluge-web --base=deluge $(/usr/bin/deluge-web --version | grep deluge-web | cut -f2 -d' ' | grep -q '^1.' && echo '' || echo '--do-not-daemonize')"
Restart=on-failure
User=debian-deluged
Group=debian-deluged
LockPersonality=yes
NoNewPrivileges=yes
PrivateDevices=yes
PrivateTmp=yes
ProtectControlGroups=yes
ProtectKernelLogs=yes
ProtectKernelModules=yes
ProtectKernelTunables=yes
ProtectSystem=yes
RestrictAddressFamilies=AF_UNIX AF_INET AF_INET6
RestrictRealtime=yes
StateDirectory=deluged
SystemCallArchitectures=native

[Install]
WantedBy=multi-user.target
'''  # noqa: E501


def _set_configuration(filename, parameter, value):
    """Set the configuration parameter."""
    deluged_is_running = action_utils.service_is_running('deluged')
    if deluged_is_running:
        action_utils.service_stop('deluged')
    deluge_web_is_running = action_utils.service_is_running('deluge-web')
    if deluge_web_is_running:
        action_utils.service_stop('deluge-web')

    with Config(DELUGE_CONF_DIR / filename) as config:
        config.content[parameter] = value

    if deluged_is_running:
        action_utils.service_start('deluged')
    if deluge_web_is_running:
        action_utils.service_start('deluge-web')


def _get_host_id():
    """Get default host id."""
    try:
        with Config(DELUGE_CONF_DIR / 'hostlist.conf') as config:
            return config.content['hosts'][0][0]
    except FileNotFoundError:
        with Config(DELUGE_CONF_DIR / 'hostlist.conf.1.2') as config:
            return config.content['hosts'][0][0]


def _set_deluged_daemon_options():
    """Set deluged daemon options."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Shellvars/lens', 'Shellvars.lns')
    aug.set('/augeas/load/Shellvars/incl[last() + 1]', DELUGED_DEFAULT_FILE)
    aug.load()
    aug.set('/files' + DELUGED_DEFAULT_FILE + '/ENABLE_DELUGED', '1')
    # overwrite daemon args, use default config directory (same as deluge-web)
    aug.set('/files' + DELUGED_DEFAULT_FILE + '/DAEMON_ARGS',
            '"-d -l /var/log/deluged/daemon.log -L info"')
    aug.save()


@privileged
def get_configuration() -> dict[str, str]:
    """Return the current deluged configuration."""
    with Config(DELUGE_CONF_DIR / 'core.conf') as config:
        download_location = config.content['download_location']

    return {'download_location': download_location}


@privileged
def set_configuration(download_location: str):
    """Set the deluged configuration."""
    _set_configuration('core.conf', 'download_location', download_location)


@privileged
def setup():
    """Perform initial setup for deluge."""
    with open(DELUGE_WEB_SYSTEMD_SERVICE_PATH, 'w',
              encoding='utf-8') as file_handle:
        file_handle.write(DELUGE_WEB_SYSTEMD_SERVICE)

    _set_deluged_daemon_options()

    subprocess.check_call(['systemctl', 'daemon-reload'])

    # Restarting an old deluge-web service stops also possible deluged process
    # that was started from the web interface.
    action_utils.service_try_restart('deluge-web')

    # Wait until core configuration is available so that status of the app can
    # be shown properly.
    _wait_for_configuration('deluged', 'core.conf')

    # Configure deluge-web to autoconnect to the default deluged daemon.
    _wait_for_configuration('deluge-web', 'web.conf')
    host_id = _get_host_id()
    _set_configuration('web.conf', 'default_daemon', host_id)


def _wait_for_configuration(service, file_name):
    """Wait until configuration file has been created."""
    conf_file = DELUGE_CONF_DIR / file_name
    if conf_file.exists():
        return

    # deluge-web creates files on first run. deluged on the other than differs
    # in version. Older version in Debian Buster creates the files after a
    # restart while newer versions create the files on first run. The following
    # approach is slightly better for create-on-exit case.
    is_running = action_utils.service_is_running(service)
    for interval in range(7):
        action_utils.service_restart(service)
        if conf_file.exists():
            break

        print('Waiting for {service} configuration')
        time.sleep(2**interval)  # Exponentially increase the time waited
    else:
        raise Exception(f'Unable to setup {service}.')

    if not is_running:
        action_utils.service_stop(service)


@privileged
def uninstall():
    """Remove configuration and service files."""
    # /etc/default/deluged is removed on purge
    pathlib.Path(DELUGE_WEB_SYSTEMD_SERVICE_PATH).unlink(missing_ok=True)
    shutil.rmtree(DELUGE_CONF_DIR, ignore_errors=True)



# SPDX-License-Identifier: AGPL-3.0-or-later

from plinth.errors import PlinthError


class BorgError(PlinthError):
    """Generic borg errors"""


class BorgRepositoryDoesNotExistError(BorgError):
    """Borg access to a repository works but the repository does not exist"""


class SshfsError(PlinthError):
    """Generic sshfs errors"""


class BorgRepositoryExists(BorgError):
    """A repository at target location already exists during initialization."""


class BorgUnencryptedRepository(BorgError):
    """Attempt to provide password on an unencrypted repository."""



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for help app.
"""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.system, pytest.mark.essential, pytest.mark.help]

# TODO Scenario: Visit the wiki
# TODO Scenario: Visit the mailing list
# TODO Scenario: Visit the IRC channel
# TODO Scenario: View the manual
# TODO Scenario: View the about page


@pytest.fixture(scope='module', autouse=True)
def fixture_background(session_browser):
    """Login."""
    functional.login(session_browser)


def test_view_status_logs(session_browser):
    """Test viewing the status logs."""
    _go_to_status_logs(session_browser)
    assert _are_status_logs_shown(session_browser)


def _go_to_status_logs(browser):
    functional.visit(browser, '/plinth/help/status-log/')


def _are_status_logs_shown(browser):
    status_log = browser.find_by_css('.status-log').first.text
    return ('-- No entries --' in status_log
            or status_log.strip().splitlines())



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the I2P module.
"""

from django.urls import re_path

from plinth.modules.i2p import views

urlpatterns = [
    re_path(r'^apps/i2p/$', views.I2PAppView.as_view(), name='index')
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for the UsersAndGroups app component.
"""

import pytest

from ..components import UsersAndGroups


@pytest.fixture(autouse=True)
def fixture_empty_components():
    """Remove all components from the global list before every test."""
    UsersAndGroups._all_components = set()


def test_create_users_and_groups_component():
    """Test initialization of users and groups component."""
    component = UsersAndGroups('simple-component')
    assert component.groups == {}
    assert component.reserved_usernames == []
    assert len(component._all_components) == 1
    assert component in component._all_components

    groups = {'test-group1', 'Test description'}
    component = UsersAndGroups('another-component', groups=groups,
                               reserved_usernames=['test-user1'])
    assert component.groups == groups
    assert component.reserved_usernames == ['test-user1']
    assert len(component._all_components) == 2
    assert component in component._all_components


def test_get_groups():
    """Test getting all the groups.

    Test that:
    1. Group names are unique
    2. All components have the same global set of groups

    """
    UsersAndGroups('component-with-no-groups')
    UsersAndGroups('component-with-one-group',
                   groups={'group1': 'description1'})
    UsersAndGroups('component-with-groups', groups={
        'group1': 'description1',
        'group2': 'description2'
    })

    assert UsersAndGroups.get_groups() == {'group1', 'group2'}
    assert UsersAndGroups.get_group_choices() == [
        ('group1', 'description1 (group1)'),
        ('group2', 'description2 (group2)')
    ]


def test_check_username_reservation():
    """Test username reservations by multiple components."""
    UsersAndGroups('complex-component',
                   reserved_usernames=['username1', 'username2'],
                   groups={'somegroup', 'some description'})
    assert not UsersAndGroups.is_username_reserved('something')
    assert UsersAndGroups.is_username_reserved('username1')

    assert not UsersAndGroups.is_username_reserved('username3')
    UsersAndGroups('temp-component', reserved_usernames=['username3'])
    assert UsersAndGroups.is_username_reserved('username3')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for Coturn daemon."""

import pathlib
import random
import shutil
import string

import augeas

from plinth import action_utils
from plinth.actions import privileged

CONFIG_FILE = pathlib.Path('/etc/coturn/freedombox.conf')


def _key_path(key):
    """Return the augeas path for a key."""
    return '/files' + str(CONFIG_FILE) + '/' + key


@privileged
def setup():
    """Setup Coturn server."""
    CONFIG_FILE.parent.mkdir(exist_ok=True)
    if not CONFIG_FILE.exists():
        CONFIG_FILE.touch(0o640)
        shutil.chown(CONFIG_FILE, group='turnserver')

    action_utils.service_daemon_reload()

    aug = augeas_load()

    # XXX: Should we set external-ip
    aug.set(_key_path('min-port'), '49152')
    aug.set(_key_path('max-port'), '50175')
    aug.set(_key_path('use-auth-secret'), 'true')
    if not aug.get(_key_path('static-auth-secret')):
        secret = ''.join(
            random.choice(string.ascii_letters + string.digits)
            for _ in range(64))
        aug.set(_key_path('static-auth-secret'), secret)

    aug.set(_key_path('cert'), '/etc/coturn/certs/cert.pem')
    aug.set(_key_path('pkey'), '/etc/coturn/certs/pkey.pem')
    aug.set(_key_path('no-tlsv1'), 'true')
    aug.set(_key_path('no-tlsv1_1'), 'true')
    aug.set(_key_path('no-cli'), 'true')
    aug.set(_key_path('listening-ip[1]'), '::')
    # Keep ::1 because at least two IP addresses of same class are needed for
    # enabling alternate port (port + 1). This is in turn needed for NAT
    # Behavior Discovery (RFC 5780).
    aug.set(_key_path('listening-ip[2]'), '::1')

    aug.save()

    action_utils.service_try_restart('coturn')


@privileged
def get_config() -> dict[str, str]:
    """Return the current configuration."""
    aug = augeas_load()
    config = {
        'static_auth_secret': aug.get(_key_path('static-auth-secret')),
        'realm': aug.get(_key_path('realm')),
    }
    return config


@privileged
def set_domain(domain_name: str):
    """Set the TLS domain.

    This value is usually not stored. So, set realm value even though it is not
    needed to set realm for REST API based authentication.
    """
    aug = augeas_load()
    aug.set(_key_path('realm'), domain_name)
    aug.save()


def augeas_load():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Simplevars/lens', 'Simplevars.lns')
    aug.set('/augeas/load/Simplevars/incl[last() + 1]', str(CONFIG_FILE))
    aug.load()

    return aug


@privileged
def uninstall():
    """Remove configuration file."""
    CONFIG_FILE.unlink(missing_ok=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Wrapper to handle package installation with apt-get."""

import logging
import os
import subprocess
from collections import defaultdict
from typing import Any

import apt.cache
import apt_inst
import apt_pkg

from plinth import action_utils
from plinth import app as app_module
from plinth import module_loader
from plinth.action_utils import run_apt_command
from plinth.actions import privileged

logger = logging.getLogger(__name__)


@privileged
def update():
    """Update apt package lists."""
    returncode = run_apt_command(['update'])
    if returncode:
        raise RuntimeError(
            f'Apt command failed with return code: {returncode}')


@privileged
def install(app_id: str, packages: list[str], skip_recommends: bool = False,
            force_configuration: str | None = None, reinstall: bool = False,
            force_missing_configuration: bool = False):
    """Install packages using apt-get."""
    if force_configuration not in ('old', 'new', None):
        raise ValueError('Invalid value for force_configuration')

    try:
        _assert_managed_packages(app_id, packages)
    except Exception:
        raise PermissionError(f'Packages are not managed: {packages}')

    extra_arguments = []
    if skip_recommends:
        extra_arguments.append('--no-install-recommends')

    if force_configuration == 'old':
        extra_arguments += [
            '-o', 'Dpkg::Options::=--force-confdef', '-o',
            'Dpkg::Options::=--force-confold'
        ]
    elif force_configuration == 'new':
        extra_arguments += ['-o', 'Dpkg::Options::=--force-confnew']

    if reinstall:
        extra_arguments.append('--reinstall')

    if force_missing_configuration:
        extra_arguments += ['-o', 'Dpkg::Options::=--force-confmiss']

    subprocess.run(['dpkg', '--configure', '-a'], check=False)
    with action_utils.apt_hold_freedombox():
        run_apt_command(['--fix-broken', 'install'])
        returncode = run_apt_command(['install'] + extra_arguments + packages)

    if returncode:
        raise RuntimeError(
            f'Apt command failed with return code: {returncode}')


@privileged
def remove(app_id: str, packages: list[str], purge: bool):
    """Remove packages using apt-get."""
    try:
        _assert_managed_packages(app_id, packages)
    except Exception:
        raise PermissionError(f'Packages are not managed: {packages}')

    subprocess.run(['dpkg', '--configure', '-a'], check=False)
    with action_utils.apt_hold_freedombox():
        run_apt_command(['--fix-broken', 'install'])
        options = [] if not purge else ['--purge']
        returncode = run_apt_command(['remove'] + options + packages)

    if returncode:
        raise RuntimeError(
            f'Apt command failed with return code: {returncode}')


def _assert_managed_packages(app_id, packages):
    """Check that list of packages are in fact managed by module."""
    from plinth.package import Packages

    module_loader.load_modules()
    app_module.apps_init()
    app = app_module.App.get(app_id)
    managed_packages = []
    for component in app.get_components_of_type(Packages):
        managed_packages += component.possible_packages + component.conflicts

    for package in packages:
        assert package in managed_packages


@privileged
def is_package_manager_busy() -> bool:
    """Check whether package manager is busy.

    An exit code of zero indicates that package manager is busy.
    """
    return action_utils.is_package_manager_busy()


@privileged
def filter_conffile_packages(
        packages_list: list[str]) -> dict[str, dict[str, Any]]:
    """Return filtered list of packages which have pending conffile prompts.

    When considering which file needs a configuration file prompt, mimic the
    behavior of unattended-upgrades package. This is because when
    unattended-upgrades gives up on the package due to conffile prompts, that
    is when this logic needs to step in and perform the upgrades.

    The logic is (roughly):

      - Read /var/lib/dpkg/status file to read hashes as provided by currently
        installed version of a package.

      - Read each configuration file for the package from disk and compute
        hashes.

      - If the hashes match, package has no configuration file that got
        modified. There will be no conffile prompt.

      - If hashes don't match then check if the version of the package being
        upgraded to has the same hash as the old version of the package or in
        the new version or the package that configuration file has been
        removed. If the conditions satisfy, then there will be no conffile
        prompt. Otherwise, package will have conffile prompt.

      - A further condition for showing conffile prompt is when new package
        brings in additional configuration files not known before and some of
        which are already present on the disk and mismatch with incoming files.

    """
    apt_pkg.init()  # Read configuration that will be used later.
    packages = set(packages_list)

    status_hashes, current_versions = _get_conffile_hashes_from_status_file(
        packages)

    mismatched_hashes = _filter_matching_package_hashes(status_hashes)

    downloaded_files = _download_packages(packages)

    new_package_hashes, new_versions = \
        _get_conffile_hashes_from_downloaded_files(
            packages, downloaded_files, status_hashes, mismatched_hashes)

    packages_info = {}
    for package in packages:
        modified_conffiles = _get_modified_conffiles(
            status_hashes[package], mismatched_hashes[package],
            new_package_hashes[package])
        if not modified_conffiles:
            continue

        package_info = {
            'current_version': current_versions[package],
            'new_version': new_versions[package],
            'modified_conffiles': modified_conffiles
        }
        packages_info[package] = package_info

    return packages_info


def _get_modified_conffiles(status_hashes, mismatched_hashes,
                            new_package_hashes):
    """Return list of conffiles that will cause prompts for a package."""
    modified_conffiles = []
    for conffile, hash_value in mismatched_hashes.items():
        if conffile not in new_package_hashes:
            # Modified configuration file not present new package
            continue

        if status_hashes[conffile] == new_package_hashes[conffile]:
            # Configuration file is same in both versions of package. Conffile
            # prompt is not triggered even if the file is modified on disk.
            continue

        modified_conffiles.append(conffile)

    for conffile, hash_value in new_package_hashes.items():
        if conffile in status_hashes:
            # File is not new, it was read and matched against new/old hashes
            continue

        if not os.path.exists(conffile):
            # New configuration file brought by new package doesn't not already
            # exist on disk.
            continue

        if _get_conffile_hash(conffile) != hash_value:
            # New configuration file brought by new package doesn't match file
            # on the disk.
            #
            # If existing file is a directory, unattended-upgrades allows it,
            # we still treat it as a conffile prompt. This should be okay.
            modified_conffiles.append(conffile)

    return modified_conffiles


def _get_conffile_hashes_from_status_file(packages):
    """For each of the packages, return a dict of conffile hashes.

    Work on all packages at the same time to avoid parsing the status file
    multiple times.

    """
    package_hashes = defaultdict(dict)
    package_versions = defaultdict(lambda: None)

    status_file = apt_pkg.config.find('Dir::State::status')
    with apt_pkg.TagFile(status_file) as tag_file:
        for section in tag_file:
            package = section.get('Package')
            if package not in packages:
                continue

            hashes = _parse_conffiles_value(section.get('Conffiles', ''))
            package_hashes[package] = hashes
            package_versions[package] = section.get('Version')

    return package_hashes, package_versions


def _parse_conffiles_value(value):
    """Parse and return the list of conffiles as found in dpkg status file."""
    conffiles = {}
    for line in value.splitlines():
        parts = line.strip().split()
        if len(parts) > 2 and parts[2] == 'obsolete':
            continue

        md5sum = parts[1]
        if md5sum == 'newconffile':  # (LP: #936870)
            continue

        file_path = parts[0]
        conffiles[file_path] = md5sum

    return conffiles


def _filter_matching_package_hashes(package_hashes):
    """Return hashes of only conffiles that don't match for each package."""
    mismatched_hashes = defaultdict(dict)
    for package, hashes in package_hashes.items():
        system_hashes = {}
        for conffile, md5sum in hashes.items():
            system_md5sum = _get_conffile_hash(conffile)
            if md5sum != system_md5sum:
                system_hashes[conffile] = system_md5sum

        if system_hashes:
            mismatched_hashes[package] = system_hashes

    return mismatched_hashes


def _get_conffile_hash(conffile):
    """Return hash of a conffile in the system."""
    try:
        with open(conffile, 'rb') as file_handle:
            return apt_pkg.md5sum(file_handle)
    except (FileNotFoundError, OSError):
        return None


def _download_packages(packages):
    """Download the package for upgrade."""
    sources_list = apt_pkg.SourceList()
    sources_list.read_main_list()

    apt_pkg_cache = apt_pkg.Cache(None)  # None prevents progress messages
    apt_cache = apt.cache.Cache()
    dep_cache = apt_pkg.DepCache(apt_pkg_cache)
    for package_name in packages:
        package = apt_cache[package_name]
        if package.is_upgradable:
            dep_cache.mark_install(apt_pkg_cache[package_name], True,
                                   not package.is_auto_installed)

    package_manager = apt_pkg.PackageManager(dep_cache)
    records = apt_pkg.PackageRecords(apt_pkg_cache)
    fetcher = apt_pkg.Acquire()
    package_manager.get_archives(fetcher, sources_list, records)
    run_result = fetcher.run()
    if run_result != apt_pkg.Acquire.RESULT_CONTINUE:
        logger.error('Downloading packages failed.')
        raise RuntimeError('Downloading packages failed.')

    downloaded_files = []
    for item in fetcher.items:
        if not item.complete or item.status == item.STAT_ERROR or \
           item.status == item.STAT_AUTH_ERROR:
            continue

        if not item.is_trusted:
            continue

        if not os.path.exists(item.destfile):
            continue

        if not item.destfile.endswith('.deb'):
            continue

        downloaded_files.append(item.destfile)

    return downloaded_files


def _get_conffile_hashes_from_downloaded_files(packages, downloaded_files,
                                               status_hashes,
                                               mismatched_hashes):
    """Retrieve the conffile hashes from downloaded .deb files."""
    new_hashes = defaultdict(dict)
    new_versions = defaultdict(lambda: None)
    for downloaded_file in downloaded_files:
        try:
            package_name, hashes, new_version = \
                _get_conffile_hashes_from_downloaded_file(
                    packages, downloaded_file, status_hashes,
                    mismatched_hashes)
        except (LookupError, apt_pkg.Error, ValueError):
            continue

        new_hashes[package_name] = hashes
        new_versions[package_name] = new_version

    return new_hashes, new_versions


def _get_conffile_hashes_from_downloaded_file(packages, downloaded_file,
                                              status_hashes,
                                              mismatched_hashes):
    """Retrieve the conffile hashes from a single downloaded .deb file."""
    deb_file = apt_inst.DebFile(downloaded_file)

    control = deb_file.control.extractdata('control')

    section = apt_pkg.TagSection(control)
    package_name = section['Package']
    if package_name not in packages:
        raise ValueError

    new_version = section['Version']

    conffiles = deb_file.control.extractdata('conffiles')
    conffiles = conffiles.decode().strip().split()

    status_hashes = status_hashes.get(package_name, {})
    mismatched_hashes = mismatched_hashes.get(package_name, {})

    hashes = {}
    for conffile in conffiles:
        if conffile in status_hashes and conffile not in mismatched_hashes:
            # File already in old package and there is no change on disk.
            # Optimization to make sure we read as fewer files as possible.
            continue

        conffile_data = deb_file.data.extractdata(conffile.lstrip('/'))
        md5sum = apt_pkg.md5sum(conffile_data)
        hashes[conffile] = md5sum

    return package_name, hashes, new_version



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for the bepasty app."""

from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.decorators.http import require_POST
from django.views.generic import FormView

from plinth.views import AppView

from . import privileged
from .forms import AddPasswordForm, SetDefaultPermissionsForm

# i18n for permission comments
PERMISSION_COMMENTS_STRINGS = {
    'admin': _('admin'),
    'editor': _('editor'),
    'viewer': _('viewer'),
}


class BepastyView(AppView):
    """Serve configuration page."""

    app_id = 'bepasty'
    form_class = SetDefaultPermissionsForm
    template_name = 'bepasty.html'

    def __init__(self, *args, **kwargs):
        """Initialize the view."""
        super().__init__(*args, **kwargs)
        self.conf = None

    def _get_configuration(self):
        """Return the current configuration."""
        if not self.conf:
            self.conf = privileged.get_configuration()

        return self.conf

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        permissions_short_text = {
            'read': _('Read'),
            'create': _('Create'),
            'list': _('List'),
            'delete': _('Delete'),
            'admin': _('Admin')
        }
        context = super().get_context_data(**kwargs)
        conf = self._get_configuration()
        passwords = []
        for password, permissions in conf['PERMISSIONS'].items():
            permissions = permissions.split(',')
            permissions = [
                str(permissions_short_text[permission])
                for permission in permissions if permission
            ]
            passwords.append({
                'password': password,
                'permissions': ', '.join(permissions),
                'comment': conf['PERMISSION_COMMENTS'].get(password) or ''
            })
        context['passwords'] = passwords
        return context

    def get_initial(self):
        """Return the status of the service to fill in the form."""
        initial = super().get_initial()
        default = self._get_configuration().get('DEFAULT_PERMISSIONS', '')
        default = ' '.join(default.split(','))
        default = 'read list' if default == 'list read' else default
        initial['default_permissions'] = default
        return initial

    def form_valid(self, form):
        """Apply the changes submitted in the form."""
        old_data = form.initial
        form_data = form.cleaned_data

        if old_data['default_permissions'] != form_data['default_permissions']:
            try:
                privileged.set_default(
                    form_data['default_permissions'].split(' '))
                messages.success(self.request, _('Configuration updated.'))
            except Exception:
                messages.error(self.request,
                               _('An error occurred during configuration.'))

        return super().form_valid(form)


class AddPasswordView(SuccessMessageMixin, FormView):
    """View to add a new password."""

    form_class = AddPasswordForm
    prefix = 'bepasty'
    template_name = 'form.html'
    success_url = reverse_lazy('bepasty:index')
    success_message = _('Password added.')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Add Password')
        return context

    def form_valid(self, form):
        """Add the password on valid form submission."""
        form_data = form.cleaned_data
        privileged.add_password(form_data['permissions'], form_data['comment'])
        return super().form_valid(form)


@require_POST
def remove(request, password):
    """View to remove a password."""
    privileged.remove_password(password)
    messages.success(request, _('Password deleted.'))
    return redirect(reverse_lazy('bepasty:index'))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for cockpit app.
"""

import pytest

from plinth.tests.functional import BaseAppTests

pytestmark = [pytest.mark.system, pytest.mark.essential, pytest.mark.cockpit]


class TestCockpitApp(BaseAppTests):
    app_name = 'cockpit'
    has_service = True
    has_web = True



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the janus app.
"""

from django.urls import re_path
from stronghold.decorators import public

from plinth.views import AppView

from .views import JanusRoomView

urlpatterns = [
    re_path(r'^apps/janus/$', AppView.as_view(app_id='janus'), name='index'),
    re_path(r'^apps/janus/room/$', public(JanusRoomView.as_view()),
            name='room')
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the Tor Proxy module.
"""

from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^apps/torproxy/$', views.TorProxyAppView.as_view(),
            name='index'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for Notification, API to show notificatoins.
"""

import datetime
from unittest.mock import patch

import pytest
from django.contrib.auth.models import Group, User
from django.core.exceptions import ValidationError

from plinth.notification import Notification

pytestmark = pytest.mark.django_db


@pytest.fixture(name='note')
def fixture_note():
    """Fixture to return a valid notification object."""
    Notification.objects.all().delete()
    return Notification.update_or_create(id='test-notification',
                                         app_id='test-app', severity='info',
                                         title='Test Title',
                                         data={'test-key': 'test-value'})


@pytest.fixture(name='user')
def fixture_user():
    """Fixture to create a new user for tests."""
    group = Group(name='test-group')
    group.save()

    user = User(username='test-user')
    user.save()
    user.groups.add(group)
    user.save()

    return user


@pytest.mark.parametrize('severity,severity_value', [('exception', 5),
                                                     ('error', 4),
                                                     ('warning', 3),
                                                     ('info', 2), ('debug', 1),
                                                     ('x-invalid', 2)])
def test_severity_value(note, severity, severity_value):
    """Test that severity_value is mapper properly."""
    note.severity = severity
    assert note.severity_value == severity_value


def test_dismiss(note):
    """Test that setting dismissed value works."""
    note.dismiss()
    assert note.dismissed
    Notification.get('test-notification') == note  # Saved

    note.dismiss(should_dismiss=True)
    assert note.dismissed

    note.dismiss(should_dismiss=False)
    assert not note.dismissed


def test_severity_invalid_value(note):
    """Test valid values for severity."""
    note.severity = 'x-invalid-value'
    with pytest.raises(ValidationError):
        note.clean()


@pytest.mark.parametrize('severity',
                         ['exception', 'error', 'warning', 'info', 'debug'])
def test_severity_valid_values(note, severity):
    """Test that valid values are accepted for severity."""
    note.severity = severity
    note.clean()


@pytest.mark.parametrize('values', [
    [None, None, None],
    ['Message', None, None],
    [None, [{
        'type': 'dismiss'
    }], None],
    [None, None, 'test-template.html'],
])
def test_valid_message_actions_body_template(note, values):
    """Test valid values for message/actions/body_template."""
    note.message = values[0]
    note.actions = values[1]
    note.body_template = values[2]
    note.clean()


@pytest.mark.parametrize('values', [
    ['Message', None, 'test-template.html'],
    [None, [{
        'type': 'dismiss'
    }], 'test-template.html'],
])
def test_invalid_message_actions_body_template(note, values):
    """Test invalid values for message/actions/body_template."""
    note.message = values[0]
    note.actions = values[1]
    note.body_template = values[2]
    with pytest.raises(ValidationError):
        note.clean()


@pytest.mark.parametrize('actions', [
    None,
    [],
    [{
        'type': 'dismiss'
    }],
    [{
        'type': 'dismiss'
    }, {
        'type': 'dismiss'
    }],
    [{
        'type': 'dismiss',
        'extra-key': 'extra-value'
    }],
    [{
        'type': 'link',
        'text': 'Test',
        'url': 'test-url'
    }],
    [{
        'type': 'link',
        'text': 'Test',
        'url': 'test-url',
        'class': 'primary'
    }],
    [{
        'type': 'link',
        'text': 'Test',
        'url': 'test-url',
        'class': 'default'
    }],
    [{
        'type': 'link',
        'text': 'Test',
        'url': 'test-url',
        'class': 'warning'
    }],
    [{
        'type': 'link',
        'text': 'Test',
        'url': 'test-url',
        'class': 'danger'
    }],
    [{
        'type': 'link',
        'text': 'Test',
        'url': 'test-url',
        'class': 'success'
    }],
    [{
        'type': 'link',
        'text': 'Test',
        'url': 'test-url',
        'class': 'info'
    }],
])
def test_valid_actions(note, actions):
    """Test valid values for actions."""
    note.actions = actions
    note.clean()


@pytest.mark.parametrize('actions', [
    'actions',
    [[]],
    [None],
    ['action'],
    [{}],
    [{
        'test-key': 'test-value'
    }],
    [{
        'type': 'invalid-type'
    }],
    [{
        'type': 'link'
    }],
    [{
        'type': 'link',
        'text': 'Test'
    }],
    [{
        'type': 'link',
        'url': 'test-url'
    }],
    [{
        'type': 'link',
        'url': 'test-url',
        'text': 'Test',
        'class': 'invalid-class'
    }],
])
def test_invalid_actions(note, actions):
    """Test invalid values for actions."""
    note.actions = actions
    with pytest.raises(ValidationError):
        note.clean()


def test_update(note):
    """Test updating a existing notification."""
    note = Notification.get('test-notification')
    assert note.app_id == 'test-app'
    assert note.severity == 'info'
    assert note.title == 'Test Title'
    assert note.data == {'test-key': 'test-value'}

    Notification.update_or_create(id='test-notification', app_id='test-app2',
                                  severity='error', title='Test Title2',
                                  data={'test-key2': 'test-value2'})
    note = Notification.get('test-notification')
    assert note.app_id == 'test-app2'
    assert note.severity == 'error'
    assert note.title == 'Test Title2'
    assert note.data == {'test-key2': 'test-value2'}


def test_create(note):
    """Test creating a new notification works."""
    note.delete()
    note.update_or_create(id='test-notification', app_id='test-app2',
                          severity='error', title='Test Title2',
                          data={'test-key2': 'test-value2'})
    note = Notification.get('test-notification')
    assert note.app_id == 'test-app2'
    assert note.severity == 'error'
    assert note.title == 'Test Title2'
    assert note.data == {'test-key2': 'test-value2'}


def test_delete(note):
    """Test deleting a new notification works."""
    note.delete()
    with pytest.raises(KeyError):
        Notification.get('test-notification')


def test_get(note):
    """Test retrieving a notification."""
    assert Notification.get('test-notification') == note

    note.delete()
    with pytest.raises(KeyError):
        Notification.get('test-notification')


def test_list_filter_key(note):
    """Test that list filter with key works."""
    assert list(Notification.list(key='invalid')) == []
    assert list(Notification.list(key='test-notification')) == [note]
    assert list(Notification.list()) == [note]


def test_list_filter_app_id(note):
    """Test that list filter with app_id works."""
    assert list(Notification.list(app_id='invalid')) == []
    assert list(Notification.list(app_id='test-app')) == [note]
    assert list(Notification.list()) == [note]


def test_list_filter_dismissed(note):
    """Test that list filter with dismissed works."""
    assert list(Notification.list(dismissed=True)) == []
    assert list(Notification.list(dismissed=False)) == [note]
    assert list(Notification.list(dismissed=None)) == [note]
    note.dismiss()
    assert list(Notification.list(dismissed=True)) == [note]
    assert list(Notification.list(dismissed=False)) == []
    assert list(Notification.list(dismissed=None)) == [note]


def test_list_filter_user(note, user):
    """Test that list filter with user works."""
    # Invalid user set on notification
    note.user = 'invalid-user'
    note.save()
    assert list(Notification.list(user=user)) == []

    # Valid user set on notification
    note.user = 'test-user'
    note.save()
    assert list(Notification.list(user=user)) == [note]


def test_list_filter_group(note, user):
    """Test that list filter with group works."""
    # Invalid group set on notification
    note.group = 'invalid-group'
    note.save()
    assert list(Notification.list(user=user)) == []

    # Valid group set on notification
    note.group = 'test-group'
    note.save()
    assert list(Notification.list(user=user)) == [note]


def test_list_filter_user_and_group(note, user):
    """Test that list filter with group works."""
    # No user/group filter
    assert list(Notification.list()) == [note]

    # No user/group set on notification
    assert list(Notification.list(user=user)) == [note]

    # Invalid user set on notification
    note.user = 'invalid-user'
    note.group = 'test-group'
    note.save()
    assert list(Notification.list(user=user)) == []

    # Invalid group set on notification
    note.user = 'test-user'
    note.group = 'invalid-group'
    note.save()
    assert list(Notification.list(user=user)) == []

    # Valid user and group set on notification
    note.user = 'test-user'
    note.group = 'test-group'
    note.save()
    assert list(Notification.list(user=user)) == [note]


@patch('plinth.notification.gettext')
def test_display_context(gettext, note, user, rf):
    """Test display context for a notification."""
    request = rf.get('/plinth/help/about/')

    data = {
        'test-key1': 'test-value1',
        'test-key2': 'translate:test-value2',
        'test-key3': {
            'test-key4': 'translate:test-value4',
            'test-key5': 'test-value5'
        }
    }
    expected_data = {
        'test-key1': 'test-value1',
        'test-key2': 'translated test-value2',
        'test-key3': {
            'test-key4': 'translated test-value4',
            'test-key5': 'test-value5'
        }
    }
    actions = [{'type': 'link', 'text': 'Test text', 'url': 'Test url'}]
    expected_actions = [{
        'type': 'link',
        'text': 'translated Test text',
        'url': 'Test url'
    }]
    gettext.side_effect = lambda string: 'translated ' + string

    note.severity = 'error'
    note.title = 'Test Title {test-key1}'
    note.message = 'Test message {test-key1}'
    note.actions = actions
    note.user = 'test-user'
    note.group = 'test-group'
    note.data = data
    note.save()

    context = Notification.get_display_context(request, user)
    assert len(context['notifications']) == 1
    assert context['max_severity'] == 'error'

    context_note = context['notifications'][0]
    assert context_note['id'] == 'test-notification'
    assert context_note['app_id'] == 'test-app'
    assert context_note['severity'] == 'error'
    assert context_note['title'] == 'translated Test Title test-value1'
    assert context_note['message'] == 'translated Test message test-value1'
    assert context_note['body'] is None
    assert context_note['actions'] == expected_actions
    assert context_note['data'] == expected_data
    assert note.data == data
    now = datetime.datetime.now()
    assert (now - context_note['created_time']).seconds < 60
    assert (now - context_note['last_update_time']).seconds < 60
    assert context_note['user'] == 'test-user'
    assert context_note['group'] == 'test-group'
    assert not context_note['dismissed']


def test_display_context_body_template(note, user, load_cfg, rf):
    """Test display context for a notification with body template."""
    request = rf.get('/plinth/help/about/')

    note.body_template = 'invalid-template.html'
    note.save()

    context = Notification.get_display_context(request, user)
    assert context['notifications'][0]['body'] == {
        'content': b'Template invalid-template.html does not exist.'
    }

    note.body_template = 'test-notification.html'
    note.save()

    context = Notification.get_display_context(request, user)
    context_note = context['notifications'][0]
    assert context_note['body'].content == \
        b'Test notification body /plinth/help/about/\n'



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the samba module.
"""

from django.urls import re_path

from . import views

urlpatterns = [
    re_path(r'^apps/samba/$', views.SambaAppView.as_view(), name='index'),
    re_path(r'^apps/samba/share/(?P<mount_point>[A-Za-z0-9%_.\-~]+)/$',
            views.share, name='share')
]



# SPDX-License-Identifier: AGPL-3.0-or-later

from django import forms
from django.utils.translation import gettext_lazy as _

from plinth import cfg
from plinth.modules import first_boot


class FirstbootWizardSecretForm(forms.Form):
    """Form to collect and validate the first boot wizard secret."""
    secret = forms.CharField(
        label='', help_text=_(
            'Enter the secret generated during FreedomBox installation. '
            'This secret can also be obtained by running the command "sudo '
            'cat /var/lib/plinth/firstboot-wizard-secret" on your {box_name}'.
            format(box_name=_(cfg.box_name))), required=False,
        widget=forms.PasswordInput(
            attrs={'placeholder': _('Firstboot Wizard Secret')}))

    def validate_secret(self, secret):
        """Match the secret provided by the user with the one
        generated during installation.
        """
        secret_file_path = first_boot.get_secret_file_path()
        with open(secret_file_path, encoding='utf-8') as secret_file:
            if secret != secret_file.read().strip():
                self.add_error('secret', 'Invalid secret')

    def clean(self):
        """Override clean to add form validation logic."""
        cleaned_data = super().clean()
        if first_boot.firstboot_wizard_secret_exists():
            self.validate_secret(cleaned_data.get("secret"))
        return cleaned_data



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the Users module
"""

from django.urls import re_path
from stronghold.decorators import public

from plinth.modules.sso.views import CaptchaView, SSOLoginView, logout
from plinth.utils import non_admin_view

from . import views

urlpatterns = [
    re_path(r'^sys/users/$', views.UserList.as_view(), name='index'),
    re_path(r'^sys/users/create/$', views.UserCreate.as_view(), name='create'),
    re_path(r'^sys/users/(?P<slug>[\w.@+-]+)/edit/$',
            non_admin_view(views.UserUpdate.as_view()), name='edit'),
    re_path(r'^sys/users/(?P<slug>[\w.@+-]+)/delete/$',
            views.UserDelete.as_view(), name='delete'),
    re_path(r'^sys/users/(?P<slug>[\w.@+-]+)/change_password/$',
            non_admin_view(views.UserChangePassword.as_view()),
            name='change_password'),

    # Authnz is handled by SSO
    re_path(r'^accounts/login/$', public(SSOLoginView.as_view()),
            name='login'),
    re_path(r'^accounts/logout/$', public(logout), name='logout'),
    re_path(r'^users/firstboot/$', public(views.FirstBootView.as_view()),
            name='firstboot'),
    re_path(r'accounts/login/locked/$', public(CaptchaView.as_view()),
            name='locked_out'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for common FreedomBox views.
"""

import pytest

from plinth.views import is_safe_url


@pytest.mark.parametrize('url', [
    '/plinth/login/',
    '/',
    'safe',
])
def test_is_safe_url_valid_url(url):
    """Test valid URLs for safe URL checks."""
    assert is_safe_url(url)


@pytest.mark.parametrize(
    'url',
    [
        '',
        None,
        '\\plinth',
        '///plinth',
        'https://example.com/plinth/login/',
        'https:///example.com',
        'https:///plinth/login',
        'ftp://example.com',
        'https://[aabb::ccdd',  # Invalid IPv6
    ])
def test_is_safe_url_invalid_url(url):
    """Test invalid URLs for safe URL checks."""
    assert not is_safe_url(url)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Manage email aliases stored in sqlite database."""

import contextlib
import pwd
import sqlite3
from dataclasses import dataclass

from . import privileged


@dataclass
class Alias:
    value: str
    name: str


@contextlib.contextmanager
def _get_cursor():
    """Return a DB cursor as context manager."""
    # Turn ON autocommit mode
    db_path = '/var/lib/postfix/freedombox-aliases/aliases.sqlite3'
    connection = sqlite3.connect(db_path, isolation_level=None)
    connection.row_factory = sqlite3.Row
    try:
        cursor = connection.cursor()
        yield cursor
    finally:
        connection.close()


def get(username):
    """Get all aliases of a user."""
    query = 'SELECT name, value FROM alias WHERE value=?'
    with _get_cursor() as cursor:
        rows = cursor.execute(query, (username, ))
        return [Alias(**row) for row in rows]


def exists(name):
    """Return whether alias is already taken."""
    try:
        pwd.getpwnam(name)
        return True
    except KeyError:
        pass

    with _get_cursor() as cursor:
        query = 'SELECT COUNT(*) FROM alias WHERE name=?'
        cursor.execute(query, (name, ))
        return cursor.fetchone()[0] != 0


def put(username, name):
    """Insert if not exists a new alias."""
    query = 'INSERT INTO alias (name, value) VALUES (?, ?)'
    with _get_cursor() as cursor:
        try:
            cursor.execute(query, (name, username))
        except sqlite3.IntegrityError:
            pass  # Alias exists, rare since we are already checking


def delete(username, aliases):
    """Delete a set of aliases."""
    query = 'DELETE FROM alias WHERE value=? AND name=?'
    parameter_seq = ((username, name) for name in aliases)
    with _get_cursor() as cursor:
        cursor.execute('BEGIN')
        cursor.executemany(query, parameter_seq)
        cursor.execute('COMMIT')


def first_setup():
    """Create the database file and schema inside it."""
    privileged.aliases.setup_aliases()

    # Create schema if not exists
    query = '''
BEGIN;
CREATE TABLE IF NOT EXISTS alias (
    name TEXT NOT NULL,
    value TEXT NOT NULL,
    PRIMARY KEY (name)
);
COMMIT;
'''
    with _get_cursor() as cursor:
        cursor.executescript(query)


def setup_common_aliases(username):
    """Create aliases for common mailboxes described in RFC2142.

    See: https://datatracker.ietf.org/doc/html/rfc2142
    """
    aliases = [
        'mailer-daemon', 'postmaster', 'nobody', 'webmaster', 'www',
        'hostmaster', 'info', 'support', 'abuse', 'noc', 'security', 'usenet',
        'news', 'ftp'
    ]
    for alias in aliases:
        if not exists(alias):
            put(username, alias)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure Tor."""

import json
import logging

from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop

from plinth import action_utils
from plinth import app as app_module
from plinth import cfg, kvstore, menu
from plinth import setup as setup_module_  # Not setup_module, for pytest
from plinth.daemon import (Daemon, app_is_running, diagnose_netcat,
                           diagnose_port_listening)
from plinth.diagnostic_check import DiagnosticCheck, Result
from plinth.modules import torproxy
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.modules.names.components import DomainType
from plinth.modules.torproxy.utils import is_apt_transport_tor_enabled
from plinth.modules.users.components import UsersAndGroups
from plinth.package import Packages
from plinth.signals import domain_added, domain_removed
from plinth.utils import format_lazy

from . import manifest, privileged, utils

logger = logging.getLogger(__name__)

_description = [
    _('Tor is an anonymous communication system. You can learn more '
      'about it from the <a href="https://www.torproject.org/">Tor '
      'Project</a> website. For best protection when web surfing, the '
      'Tor Project recommends that you use the '
      '<a href="https://www.torproject.org/download/download-easy.html.en">'
      'Tor Browser</a>.'),
    _('This app provides relay services to contribute to Tor network and help '
      'others overcome censorship.'),
    format_lazy(
        _('This app provides an onion domain to expose {box_name} services '
          'via the Tor network. Using Tor browser, one can access {box_name} '
          'from the internet even when using an ISP that limits servers at '
          'home.'), box_name=_(cfg.box_name)),
]


class TorApp(app_module.App):
    """FreedomBox app for Tor."""

    app_id = 'tor'

    _version = 7

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               depends=['names'
                                        ], name=_('Tor'), icon_filename='tor',
                               short_description=_('Anonymity Network'),
                               description=_description, manual_page='Tor',
                               clients=manifest.clients,
                               donation_url='https://donate.torproject.org/')
        self.add(info)

        menu_item = menu.Menu('menu-tor', info.name, info.short_description,
                              info.icon_filename, 'tor:index',
                              parent_url_name='apps')
        self.add(menu_item)

        packages = Packages('packages-tor',
                            ['tor', 'tor-geoipdb', 'obfs4proxy'])
        self.add(packages)

        domain_type = DomainType('domain-type-tor', _('Tor Onion Service'),
                                 'tor:index', can_have_certificate=False)
        self.add(domain_type)

        firewall = Firewall('firewall-tor-relay', _('Tor Bridge Relay'),
                            ports=['tor-orport', 'tor-obfs3',
                                   'tor-obfs4'], is_external=True)
        self.add(firewall)

        daemon = Daemon('daemon-tor', 'tor@plinth', strict_check=True)
        self.add(daemon)

        webserver = Webserver('webserver-onion-location',
                              'onion-location-freedombox')
        self.add(webserver)

        users_and_groups = UsersAndGroups('users-and-groups-tor',
                                          reserved_usernames=['debian-tor'])
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-tor', **manifest.backup)
        self.add(backup_restore)

    def post_init(self):
        """Perform post initialization operations."""
        # Register onion service name with Name Services module.
        if (not self.needs_setup() and self.is_enabled()
                and app_is_running(self)):
            status = utils.get_status(initialized=False)
            hostname = status['hs_hostname']
            services = [int(port['virtport']) for port in status['hs_ports']]

            if status['hs_enabled'] and status['hs_hostname']:
                domain_added.send_robust(sender='tor',
                                         domain_type='domain-type-tor',
                                         name=hostname, services=services)

    def enable(self):
        """Enable the app and update firewall ports."""
        super().enable()
        privileged.update_ports()
        update_hidden_service_domain()

    def disable(self):
        """Disable the app and remove HS domain."""
        super().disable()
        update_hidden_service_domain()

    def diagnose(self) -> list[DiagnosticCheck]:
        """Run diagnostics and return the results."""
        results = super().diagnose()

        results.extend(_diagnose_control_port())

        status = utils.get_status()
        ports = status['ports']

        if status['relay_enabled']:
            results.append(
                DiagnosticCheck(
                    'tor-port-relay', gettext_noop('Tor relay port available'),
                    Result.PASSED if 'orport' in ports else Result.FAILED))
            if 'orport' in ports:
                results.append(
                    diagnose_port_listening(int(ports['orport']), 'tcp4'))
                results.append(
                    diagnose_port_listening(int(ports['orport']), 'tcp6'))

        if status['bridge_relay_enabled']:
            results.append(
                DiagnosticCheck(
                    'tor-port-obfs3',
                    gettext_noop('Obfs3 transport registered'),
                    Result.PASSED if 'obfs3' in ports else Result.FAILED))
            if 'obfs3' in ports:
                results.append(
                    diagnose_port_listening(int(ports['obfs3']), 'tcp4'))
                results.append(
                    diagnose_port_listening(int(ports['obfs3']), 'tcp6'))

            results.append(
                DiagnosticCheck(
                    'tor-port-obfs4',
                    gettext_noop('Obfs4 transport registered'),
                    Result.PASSED if 'obfs4' in ports else Result.FAILED))
            if 'obfs4' in ports:
                results.append(
                    diagnose_port_listening(int(ports['obfs4']), 'tcp4'))
                results.append(
                    diagnose_port_listening(int(ports['obfs4']), 'tcp6'))

        if status['hs_enabled']:
            hs_hostname = status['hs_hostname'].split('.onion')[0]
            results.append(
                DiagnosticCheck(
                    'tor-onion-version',
                    gettext_noop('Onion service is version 3'), Result.PASSED
                    if len(hs_hostname) == 56 else Result.FAILED))

        return results

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup(old_version)
        status = utils.get_status()
        update_hidden_service_domain(status)

        # Enable/disable Onion-Location component based on app status.
        # Component was introduced in version 6.
        if old_version and old_version < 6:
            daemon_component = self.get_component('daemon-tor')
            component = self.get_component('webserver-onion-location')
            if daemon_component.is_enabled():
                logger.info('Enabling Onion-Location component')
                component.enable()
            else:
                logger.info('Disabling Onion-Location component')
                component.disable()

        # The SOCKS proxy and "Download software packages using Tor" features
        # were moved into a new app, Tor Proxy, in version 7. If Tor is
        # enabled, then store the relevant configuration, and install Tor
        # Proxy.
        if old_version and old_version < 7 and self.is_enabled():
            logger.info('Tor Proxy app will be installed')
            config = {
                'use_upstream_bridges': status['use_upstream_bridges'],
                'upstream_bridges': status['upstream_bridges'],
                'apt_transport_tor': is_apt_transport_tor_enabled()
            }
            kvstore.set(torproxy.PREINSTALL_CONFIG_KEY, json.dumps(config))
            # This creates the operation, which will run after the current
            # operation (Tor setup) is completed.
            setup_module_.run_setup_on_app('torproxy')

        if not old_version:
            logger.info('Enabling Tor app')
            self.enable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()


def update_hidden_service_domain(status=None):
    """Update onion service domain with Name Services module."""
    if not status:
        status = utils.get_status()

    domain_removed.send_robust(sender='tor', domain_type='domain-type-tor')

    if status['enabled'] and status['is_running'] and \
       status['hs_enabled'] and status['hs_hostname']:
        services = [int(port['virtport']) for port in status['hs_ports']]
        domain_added.send_robust(sender='tor', domain_type='domain-type-tor',
                                 name=status['hs_hostname'], services=services)


def _diagnose_control_port() -> list[DiagnosticCheck]:
    """Diagnose whether Tor control port is open on 127.0.0.1 only."""
    results = []

    addresses = action_utils.get_ip_addresses()
    for address in addresses:
        if address['kind'] != '4':
            continue

        negate = True
        if address['address'] == '127.0.0.1':
            negate = False

        results.append(
            diagnose_netcat(str(address['address']), 9051,
                            remote_input='QUIT\n', negate=negate))

    return results



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for Minetest server."""

import augeas
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.daemon import Daemon
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.modules.users.components import UsersAndGroups
from plinth.package import Package, Packages, install
from plinth.utils import Version, format_lazy

from . import manifest, privileged

_mods = [
    Package('minetest-mod-3d-armor') | Package('minetest-mod-player-3d-armor'),
    'minetest-mod-character-creator', 'minetest-mod-craftguide',
    'minetest-mod-infinite-chest', 'minetest-mod-lucky-block',
    'minetest-mod-maidroid', 'minetest-mod-mesecons',
    'minetest-mod-moreblocks', 'minetest-mod-moreores', 'minetest-mod-nether',
    'minetest-mod-pipeworks', 'minetest-mod-protector', 'minetest-mod-quartz',
    'minetest-mod-skyblock', 'minetest-mod-throwing',
    'minetest-mod-unified-inventory', 'minetest-mod-unifieddyes',
    'minetest-mod-worldedit'
]

_description = [
    format_lazy(
        _('Minetest is a multiplayer infinite-world block sandbox. This '
          'module enables the Minetest server to be run on this '
          '{box_name}, on the default port (30000). To connect to the server, '
          'a <a href="http://www.minetest.net/downloads/">Minetest client</a> '
          'is needed.'), box_name=_(cfg.box_name)),
]

CONFIG_FILE = '/etc/minetest/minetest.conf'
AUG_PATH = '/files' + CONFIG_FILE + '/.anon'


class MinetestApp(app_module.App):
    """FreedomBox app for Minetest."""

    app_id = 'minetest'

    _version = 2

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(
            app_id=self.app_id, version=self._version, name=_('Minetest'),
            icon_filename='minetest', short_description=_('Block Sandbox'),
            description=_description, manual_page='Minetest',
            clients=manifest.clients,
            donation_url='https://www.minetest.net/get-involved/#donate')
        self.add(info)

        menu_item = menu.Menu('menu-minetest', info.name,
                              info.short_description, info.icon_filename,
                              'minetest:index', parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut(
            'shortcut-minetest', info.name,
            short_description=info.short_description, icon=info.icon_filename,
            description=info.description, manual_page=info.manual_page,
            configure_url=reverse_lazy('minetest:index'), clients=info.clients,
            login_required=False)
        self.add(shortcut)

        packages = Packages('packages-minetest', ['minetest-server'] + _mods)
        self.add(packages)

        firewall = Firewall('firewall-minetest', info.name,
                            ports=['minetest-plinth'], is_external=True)
        self.add(firewall)

        daemon = Daemon('daemon-minetest', 'minetest-server',
                        listen_ports=[(30000, 'udp4')])
        self.add(daemon)

        users_and_groups = UsersAndGroups(
            'users-and-groups-minetest',
            reserved_usernames=['Debian-minetest'])
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-minetest',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        if not old_version:
            self.enable()

    def force_upgrade(self, packages):
        """Force upgrade minetest to resolve conffile prompt."""
        if 'minetest-server' not in packages:
            return False

        # Allow upgrade from 5.3.0 to 5.6.1
        package = packages['minetest-server']
        if Version(package['new_version']) > Version('5.7~'):
            return False

        config = get_configuration()
        install(['minetest-server'], force_configuration='new')
        privileged.configure(**config)

        return True


def load_augeas():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Php/lens', 'Php.lns')
    aug.set('/augeas/load/Php/incl[last() + 1]', CONFIG_FILE)
    aug.load()
    return aug


def get_max_players(aug):
    """Return the maximum players allowed on the server at one time."""
    value = aug.get(AUG_PATH + '/max_users')
    if value:
        return int(value)


def is_creative_mode_enabled(aug):
    """Return whether creative mode is enabled."""
    value = aug.get(AUG_PATH + '/creative_mode')
    return value == 'true'


def is_pvp_enabled(aug):
    """Return whether PVP is enabled."""
    value = aug.get(AUG_PATH + '/enable_pvp')
    return value == 'true'


def is_damage_enabled(aug):
    """Return whether damage is enabled."""
    value = aug.get(AUG_PATH + '/enable_damage')
    return value == 'true'


def get_configuration():
    """Return the current configuration."""
    aug = load_augeas()
    conf = {
        'max_players': get_max_players(aug),
        'creative_mode': is_creative_mode_enabled(aug),
        'enable_pvp': is_pvp_enabled(aug),
        'enable_damage': is_damage_enabled(aug),
    }
    return conf



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for Coturn app."""

from django.contrib import messages
from django.utils.translation import gettext_lazy as _

import plinth.modules.coturn as coturn
from plinth import app as app_module
from plinth import views

from . import forms


class CoturnAppView(views.AppView):
    """Serve configuration page."""

    app_id = 'coturn'
    template_name = 'coturn.html'
    form_class = forms.CoturnForm

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['config'] = coturn.get_config()
        return context

    def get_initial(self):
        """Return the values to fill in the form."""
        initial = super().get_initial()
        initial['domain'] = coturn.get_domain()
        return initial

    def form_valid(self, form):
        """Change the domain of Coturn service."""
        data = form.cleaned_data
        if coturn.get_domain() != data['domain']:
            coturn.set_domain(data['domain'])
            app = app_module.App.get('coturn')
            app.get_component('letsencrypt-coturn').setup_certificates()
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

"""
Application manifest for RSS-Bridge.
"""

clients = [{
    'name': _('RSS-Bridge'),
    'platforms': [{
        'type': 'web',
        'url': '/rss-bridge/'
    }]
}]

backup = {
        'data': {
            'files': ['/etc/rss-bridge/is_public']
        }
}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure PageKite."""

import os

import augeas

from plinth import action_utils
from plinth.actions import privileged
from plinth.modules.pagekite import utils

PATHS = {
    'service_on':
        os.path.join(utils.CONF_PATH, '*', 'service_on', '*'),
    'kitename':
        os.path.join(utils.CONF_PATH, '10_account.rc', 'kitename'),
    'kitesecret':
        os.path.join(utils.CONF_PATH, '10_account.rc', 'kitesecret'),
    'abort_not_configured':
        os.path.join(utils.CONF_PATH, '10_account.rc', 'abort_not_configured'),
    'defaults':
        os.path.join(utils.CONF_PATH, '20_frontends.rc', 'defaults'),
    'frontend':
        os.path.join(utils.CONF_PATH, '20_frontends.rc', 'frontend'),
}


@privileged
def get_config() -> dict[str, object]:
    """Return the current configuration as JSON dictionary."""
    aug = _augeas_load()
    if aug.match(PATHS['abort_not_configured']):
        aug.remove(PATHS['abort_not_configured'])
        aug.save()

    if aug.match(PATHS['defaults']):
        frontend = 'pagekite.net'
    else:
        frontend = aug.get(PATHS['frontend']) or ''

    frontend_parts = frontend.split(':')
    server_domain = frontend_parts[0]
    server_port = frontend_parts[1] if len(frontend_parts) >= 2 else '80'

    status = {
        'kite_name': aug.get(PATHS['kitename']),
        'kite_secret': aug.get(PATHS['kitesecret']),
        'server_domain': server_domain,
        'server_port': server_port,
        'predefined_services': {
            proto: False
            for proto in utils.PREDEFINED_SERVICES
        },
        'custom_services': [],
    }

    # 1. predefined_services: {'http': False, 'ssh': True, 'https': True}
    # 2. custom_services: [{'protocol': 'http', 'secret' 'nono', ..}, [..]}
    for match in aug.match(PATHS['service_on']):
        service = dict([(param, aug.get(os.path.join(match, param)))
                        for param in utils.SERVICE_PARAMS])
        for name, predefined_service in utils.PREDEFINED_SERVICES.items():
            if service == predefined_service['params']:
                status['predefined_services'][name] = True
                break
        else:
            status['custom_services'].append(service)
            if '/' in service['protocol']:
                service['protocol'], service['frontend_port'] = service[
                    'protocol'].split('/')

            service['subdomains'] = service['kitename'].startswith('*.')
            kite_name = status['kite_name']
            protocol = service['protocol']
            if service['subdomains']:
                kite_name = f'*.{kite_name}'

            url = f'{protocol}://{kite_name}'
            if 'frontend_port' in service and service['frontend_port']:
                url = "%s:%s" % (url, service['frontend_port'])

            service['url'] = url

    return status


@privileged
def set_config(frontend: str, kite_name: str, kite_secret: str):
    """Set pagekite kite name, secret and frontend URL."""
    aug = _augeas_load()
    aug.remove(PATHS['abort_not_configured'])

    aug.set(PATHS['kitename'], kite_name)
    aug.set(PATHS['kitesecret'], kite_secret)

    frontend_domain = frontend.split(':')[0]
    if frontend_domain in ('pagekite.net', 'defaults', 'default'):
        aug.set(PATHS['defaults'], '')
        aug.remove(PATHS['frontend'])
    else:
        aug.remove(PATHS['defaults'])
        aug.set(PATHS['frontend'], frontend)

    aug.save()

    for service_name in utils.PREDEFINED_SERVICES.keys():
        service = utils.PREDEFINED_SERVICES[service_name]['params']
        try:
            _add_service(aug, service)
        except RuntimeError:
            pass

    # Immediately after install, pagekite is enabled but not running. Restart
    # based on enabled state instead of try-restart.
    if action_utils.service_is_enabled('pagekite'):
        action_utils.service_restart('pagekite')


@privileged
def remove_service(service: dict[str, str]):
    """Search and remove the service(s) that match all given parameters."""
    aug = _augeas_load()
    service = utils.load_service(service)
    paths = _get_existing_service_paths(aug, service)
    # TODO: theoretically, everything to do here is:
    # [aug.remove(path) for path in paths]
    # but augeas won't let you save the changed files and doesn't say why
    for path in paths:
        filepath = _convert_augeas_path_to_filepath(path)
        service_found = False
        with open(filepath, 'r', encoding='utf-8') as file:
            lines = file.readlines()
            for i, line in enumerate(lines):
                if line.startswith('service_on') and \
                        all(param in line for param in service.values()):
                    lines[i] = ""
                    service_found = True
                    break
        if service_found:
            with open(filepath, 'w', encoding='utf-8') as file:
                file.writelines(lines)
                # abort to only allow deleting one service
                break
    action_utils.service_restart('pagekite')


def _get_existing_service_paths(aug, service, keys=None):
    """Return paths of existing services that match the given service."""
    # construct an augeas query path with patterns like:
    #     */service_on/*[protocol='http']
    path = PATHS['service_on']
    for param in (keys or service.keys()):
        path += "[%s='%s']" % (param, service[param])
    return aug.match(path)


def _add_service(aug, service):
    """Add a new service into configuration."""
    if _get_existing_service_paths(aug, service, ['protocol', 'kitename']):
        msg = "Service with the parameters %s already exists"
        raise RuntimeError(msg % service)

    root = _get_new_service_path(aug, service['protocol'])
    # TODO: after adding a service, augeas fails writing the config;
    # so add the service_on entry manually instead
    path = _convert_augeas_path_to_filepath(root)
    with open(path, 'a', encoding='utf-8') as servicefile:
        line = "\nservice_on = %s\n" % utils.convert_service_to_string(service)
        servicefile.write(line)


@privileged
def add_service(service: dict[str, str]):
    """Add one service."""
    aug = _augeas_load()
    service = utils.load_service(service)
    _add_service(aug, service)
    action_utils.service_try_restart('pagekite')


def _convert_augeas_path_to_filepath(augpath, prefix='/files',
                                     suffix='service_on'):
    """Convert an augeas service_on path to the actual file path."""
    if augpath.startswith(prefix):
        augpath = augpath.replace(prefix, "", 1)

    index = augpath.rfind(suffix)
    if index:
        augpath = augpath[:index]
    return augpath.rstrip('/')


def _get_new_service_path(aug, protocol):
    """Get the augeas path of a new service for a protocol.

    This takes care of existing services using a /service_on/*/ query
    """
    root = utils.get_augeas_servicefile_path(protocol)
    new_index = len(aug.match(root + '/*')) + 1
    return os.path.join(root, str(new_index))


def _augeas_load():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Pagekite/lens', 'Pagekite.lns')
    aug.set('/augeas/load/Pagekite/incl[last() + 1]', '/etc/pagekite.d/*.rc')
    aug.load()
    return aug



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure sharing."""

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, menu
from plinth.config import DropinConfigs
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.utils import format_lazy

from . import manifest, privileged

_description = [
    format_lazy(
        _('Sharing allows you to share files and folders on your {box_name} '
          'over the web with chosen groups of users.'),
        box_name=_(cfg.box_name))
]


class SharingApp(app_module.App):
    """FreedomBox app for sharing files."""

    app_id = 'sharing'

    _version = 3

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()
        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('Sharing'), icon_filename='sharing',
                               manual_page='Sharing', description=_description)
        self.add(info)

        menu_item = menu.Menu('menu-sharing', info.name, None,
                              info.icon_filename, 'sharing:index',
                              parent_url_name='apps')
        self.add(menu_item)

        dropin_configs = DropinConfigs('dropin-configs-sharing', [
            '/etc/apache2/includes/freedombox-sharing.conf',
        ])
        self.add(dropin_configs)

        webserver = Webserver('webserver-sharing', 'sharing-freedombox')
        self.add(webserver)

        backup_restore = BackupRestore('backup-restore-sharing',
                                       **manifest.backup)
        self.add(backup_restore)

    def has_diagnostics(self):
        """Disable diagnostics button despite having webserver component."""
        return False

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if not old_version:
            self.enable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for wireguard module forms.
"""

import pytest
from django.core.exceptions import ValidationError

from plinth.modules.wireguard.forms import validate_endpoint, validate_key


@pytest.mark.parametrize('key', [
    'gKQhVGla4UtdqeY1dQ21G5lqrnX5NFcSEAqzM5iSdl0=',
    'uHWSYIjPnS9fYFhZ0mf22IkOMyrWXDlfpXs6ve4QGHk=',
])
def test_validate_key_valid_patterns(key):
    """Test that valid wireguard key patterns as accepted."""
    validate_key(key)


@pytest.mark.parametrize(
    'key',
    [
        # Invalid padding
        'gKQhVGla4UtdqeY1dQ21G5lqrnX5NFcSEAqzM5iSdl0',
        'invalid-base64',
        '',
        'aW52YWxpZC1sZW5ndGg=',  # Incorrect length
    ])
def test_validate_key_invalid_patterns(key):
    """Test that invalid wireguard key patterns are rejected."""
    with pytest.raises(ValidationError):
        validate_key(key)


@pytest.mark.parametrize('endpoint', [
    '[1::2]:1234',
    '1.2.3.4:1234',
    'example.com:1234',
])
def test_validate_endpoint_valid_patterns(endpoint):
    """Test that valid wireguard endpoint patterns are accepted."""
    validate_endpoint(endpoint)


@pytest.mark.parametrize(
    'endpoint',
    [
        '',
        # Invalid port
        '1.2.3.4',
        '1.2.3.4:',
        '1.2.3.4:0',
        '1.2.3.4:65536',
        '1.2.3.4:1234invalid',
        '1.2.3.4:invalid',
        # Invalid IPv6
        '[]:1234',
        '[:1234',
    ])
def test_validate_endpoint_invalid_patterns(endpoint):
    """Test that invalid wireguard endpoint patterns are rejected."""
    with pytest.raises(ValidationError):
        validate_endpoint(endpoint)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Tor Proxy service."""

import logging
import os
import shutil
import subprocess
from typing import Any

import augeas

from plinth import action_utils
from plinth.actions import privileged
from plinth.modules.torproxy.utils import (APT_TOR_PREFIX, get_augeas,
                                           iter_apt_uris)

logger = logging.getLogger(__name__)

INSTANCE_NAME = 'fbxproxy'
SERVICE_FILE = '/etc/firewalld/services/tor-{0}.xml'
SERVICE_NAME = f'tor@{INSTANCE_NAME}'
TORPROXY_CONFIG = f'/etc/tor/instances/{INSTANCE_NAME}/torrc'
TORPROXY_CONFIG_AUG = f'/files/{TORPROXY_CONFIG}'


@privileged
def setup(old_version: int):
    """Setup Tor configuration after installing it."""
    _first_time_setup()


def _first_time_setup():
    """Setup Tor configuration for the first time setting defaults."""
    logger.info('Performing first time setup for Tor Proxy')
    # Disable default tor service. We will use tor@fbxproxy instance
    # instead.
    _disable_apt_transport_tor()
    action_utils.service_disable('tor@default')

    subprocess.run(['tor-instance-create', INSTANCE_NAME], check=True)

    # Remove line starting with +SocksPort, since our augeas lens
    # doesn't handle it correctly.
    with open(TORPROXY_CONFIG, 'r', encoding='utf-8') as torrc:
        torrc_lines = torrc.readlines()
    with open(TORPROXY_CONFIG, 'w', encoding='utf-8') as torrc:
        for line in torrc_lines:
            if not line.startswith('+'):
                torrc.write(line)

    aug = augeas_load()

    aug.set(TORPROXY_CONFIG_AUG + '/SocksPort[1]', '[::]:9050')
    aug.set(TORPROXY_CONFIG_AUG + '/SocksPort[2]', '0.0.0.0:9050')
    aug.set(TORPROXY_CONFIG_AUG + '/VirtualAddrNetworkIPv4', '10.192.0.0/10')
    aug.set(TORPROXY_CONFIG_AUG + '/AutomapHostsOnResolve', '1')
    aug.set(TORPROXY_CONFIG_AUG + '/TransPort[1]', '127.0.0.1:9040')
    aug.set(TORPROXY_CONFIG_AUG + '/TransPort[2]', '[::1]:9040')
    aug.set(TORPROXY_CONFIG_AUG + '/DNSPort[1]', '127.0.0.1:9053')
    aug.set(TORPROXY_CONFIG_AUG + '/DNSPort[2]', '[::1]:9053')

    aug.save()

    action_utils.service_enable(SERVICE_NAME)
    action_utils.service_restart(SERVICE_NAME)


@privileged
def configure(use_upstream_bridges: bool | None = None,
              upstream_bridges: str | None = None,
              apt_transport_tor: bool | None = None):
    """Configure Tor."""
    aug = augeas_load()

    _use_upstream_bridges(use_upstream_bridges, aug=aug)

    if upstream_bridges:
        _set_upstream_bridges(upstream_bridges, aug=aug)

    if apt_transport_tor:
        _enable_apt_transport_tor()
    elif apt_transport_tor is not None:
        _disable_apt_transport_tor()


@privileged
def restart():
    """Restart Tor."""
    if (action_utils.service_is_enabled(SERVICE_NAME, strict_check=True)
            and action_utils.service_is_running(SERVICE_NAME)):
        action_utils.service_restart(SERVICE_NAME)


@privileged
def get_status() -> dict[str, bool | str | dict[str, Any]]:
    """Return dict with Tor Proxy status."""
    aug = augeas_load()
    return {
        'use_upstream_bridges': _are_upstream_bridges_enabled(aug),
        'upstream_bridges': _get_upstream_bridges(aug)
    }


def _are_upstream_bridges_enabled(aug) -> bool:
    """Return whether upstream bridges are being used."""
    use_bridges = aug.get(TORPROXY_CONFIG_AUG + '/UseBridges')
    return use_bridges == '1'


def _get_upstream_bridges(aug) -> str:
    """Return upstream bridges separated by newlines."""
    matches = aug.match(TORPROXY_CONFIG_AUG + '/Bridge')
    bridges = [aug.get(match) for match in matches]
    return '\n'.join(bridges)


def _enable():
    """Enable and start the service."""
    action_utils.service_enable(SERVICE_NAME)


def _disable():
    """Disable and stop the service."""
    _disable_apt_transport_tor()
    action_utils.service_disable(SERVICE_NAME)


def _use_upstream_bridges(use_upstream_bridges: bool | None = None, aug=None):
    """Enable use of upstream bridges."""
    if use_upstream_bridges is None:
        return

    if not aug:
        aug = augeas_load()

    if use_upstream_bridges:
        aug.set(TORPROXY_CONFIG_AUG + '/UseBridges', '1')
    else:
        aug.set(TORPROXY_CONFIG_AUG + '/UseBridges', '0')

    aug.save()


def _set_upstream_bridges(upstream_bridges=None, aug=None):
    """Set list of upstream bridges."""
    if upstream_bridges is None:
        return

    if not aug:
        aug = augeas_load()

    aug.remove(TORPROXY_CONFIG_AUG + '/Bridge')
    if upstream_bridges:
        bridges = [bridge.strip() for bridge in upstream_bridges.split('\n')]
        bridges = [bridge for bridge in bridges if bridge]
        for bridge in bridges:
            parts = [part for part in bridge.split() if part]
            bridge = ' '.join(parts)
            aug.set(TORPROXY_CONFIG_AUG + '/Bridge[last() + 1]',
                    bridge.strip())

    aug.set(TORPROXY_CONFIG_AUG + '/ClientTransportPlugin',
            'obfs3,scramblesuit,obfs4 exec /usr/bin/obfs4proxy')

    aug.save()


def _enable_apt_transport_tor():
    """Enable package download over Tor."""
    aug = get_augeas()
    for uri_path in iter_apt_uris(aug):
        uri = aug.get(uri_path)
        if uri.startswith('http://') or uri.startswith('https://'):
            aug.set(uri_path, APT_TOR_PREFIX + uri)

    aug.save()


def _disable_apt_transport_tor():
    """Disable package download over Tor."""
    try:
        aug = get_augeas()
    except Exception:
        # Disable what we can, so APT is not unusable.
        pass

    for uri_path in iter_apt_uris(aug):
        uri = aug.get(uri_path)
        if uri.startswith(APT_TOR_PREFIX):
            aug.set(uri_path, uri[len(APT_TOR_PREFIX):])

    aug.save()


def augeas_load():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Tor/lens', 'Tor.lns')
    aug.set('/augeas/load/Tor/incl[last() + 1]', TORPROXY_CONFIG)
    aug.load()
    return aug


@privileged
def uninstall():
    """Remove fbxproxy instance."""
    directories = [
        f'/etc/tor/instances/{INSTANCE_NAME}/',
        f'/var/lib/tor-instances/{INSTANCE_NAME}/',
        f'/var/run/tor-instances/{INSTANCE_NAME}/'
    ]
    for directory in directories:
        shutil.rmtree(directory, ignore_errors=True)

    os.unlink(f'/var/run/tor-instances/{INSTANCE_NAME}.defaults')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
pytest configuration for all tests in the plinth/tests/ directory.
"""

import pathlib
from unittest.mock import patch

import pytest

from plinth import cfg


@pytest.fixture(name='shortcuts_file')
def fixture_shortcuts_file():
    with patch('plinth.frontpage.get_custom_shortcuts_paths') as func:

        def setter(file_name):
            path = pathlib.Path(__file__).parent / 'data' / 'shortcuts'
            path /= file_name
            func.return_value = cfg.expand_to_dot_d_paths([path])

        yield setter



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for Matrix Synapse STUN/TURN configuration.
"""

from unittest.mock import patch, Mock

import pytest

from plinth.modules import matrixsynapse
from plinth.modules.coturn.components import TurnConfiguration
from plinth.modules.matrixsynapse import privileged

pytestmark = pytest.mark.usefixtures('mock_privileged')
privileged_modules_to_mock = ['plinth.modules.matrixsynapse.privileged']


@pytest.fixture(name='managed_turn_conf_file')
def fixture_managed_turn_conf_file(tmp_path):
    """Returns a dummy TURN configuration file."""
    conf_file = tmp_path / 'freedombox-turn.yaml'
    return str(conf_file)


@pytest.fixture(name='overridden_turn_conf_file')
def fixture_overridden_turn_conf_file(tmp_path):
    """Returns a dummy TURN configuration file."""
    conf_file = tmp_path / 'turn.yaml'
    return str(conf_file)


@pytest.fixture(autouse=True)
def fixture_set_paths(managed_turn_conf_file, overridden_turn_conf_file):
    """Run actions with custom root path."""
    privileged.TURN_CONF_PATH = managed_turn_conf_file
    privileged.OVERRIDDEN_TURN_CONF_PATH = overridden_turn_conf_file
    with patch('plinth.privileged.service.try_restart'):
        yield


@pytest.fixture(name='test_configuration', autouse=True)
def fixture_test_configuration(managed_turn_conf_file,
                               overridden_turn_conf_file):
    """Use a separate Matrix Synapse configuration for tests.

    Overrides TURN configuration files.
    """
    matrixsynapse = 'plinth.modules.matrixsynapse'
    with (patch(f'{matrixsynapse}.privileged.TURN_CONF_PATH',
                managed_turn_conf_file),
          patch(f'{matrixsynapse}.privileged.OVERRIDDEN_TURN_CONF_PATH',
                overridden_turn_conf_file),
          patch(f'{matrixsynapse}.is_setup', return_value=True),
          patch('plinth.app.App.get') as app_get):
        app = Mock()
        app_get.return_value = app
        app.needs_setup.return_value = False
        yield


coturn_configuration = TurnConfiguration(
    'freedombox.local', [],
    'aiP02OAGyOlj6WGuCyqj7iaOsbkC7BUeKvKzhAsTZ8MEwMd3yTwpr2uvbOxgWe51')

overridden_configuration = TurnConfiguration(
    'public.coturn.site', [],
    'BUeKvKzhAsTZ8MEwMd3yTwpr2uvbOxgWe51aiP02OAGyOlj6WGuCyqj7iaOsbkC7')

updated_coturn_configuration = TurnConfiguration(
    'my.freedombox.rocks', [],
    'aiP02OsbkC7BUeKvKzhAsTZ8MEwMd3yTwpr2uvbOxgWe51AGyOlj6WGuCyqj7iaO')


def _set_managed_configuration(config=coturn_configuration):
    with patch('plinth.action_utils.service_try_restart'):
        matrixsynapse.update_turn_configuration(config)


def _set_overridden_configuration(
                                  config=overridden_configuration):
    with patch('plinth.action_utils.service_try_restart'):
        matrixsynapse.update_turn_configuration(config, managed=False)


def _assert_conf(expected_configuration, expected_managed):
    """Assert that matrix synapse TURN configuration is as expected."""
    config, managed = matrixsynapse.get_turn_configuration()
    assert config.uris == expected_configuration.uris
    assert config.shared_secret == expected_configuration.shared_secret
    assert managed == expected_managed


def test_managed_turn_server_configuration():
    """Test setting and getting managed TURN server configuration."""
    _set_managed_configuration()
    _assert_conf(coturn_configuration, True)


def test_overridden_turn_server_configuration():
    """Test setting and getting overridden TURN sever configuration."""
    _set_overridden_configuration()
    _assert_conf(overridden_configuration, False)


def test_revert_to_managed_turn_server_configuration():
    """Test setting and getting overridden TURN sever configuration."""
    # Had to do all 3 operations because all fixtures were function-scoped
    _set_managed_configuration()
    _set_overridden_configuration()
    _set_overridden_configuration(TurnConfiguration())
    _assert_conf(coturn_configuration, True)


def test_coturn_configuration_update_after_admin_override():
    """Test that overridden conf prevails even if managed conf is updated."""
    _set_managed_configuration()
    _set_overridden_configuration()
    _set_managed_configuration(updated_coturn_configuration)
    _assert_conf(overridden_configuration, False)



# SPDX-License-Identifier: AGPL-3.0-or-later

import os
from urllib.parse import urlparse

from django import template

from plinth import clients as clients_module

register = template.Library()


def mark_active_menuitem(menu, path):
    """Mark the best-matching menu item with 'active'

    Input: a menu dict in the form of:
        [{'url': '/path/to/choice1/', 'text': 'choice 1'}, {'url': ...}]

    URL paths are expected to end with a slash for matches to work properly.

    Output: The same dictionary; the best-matching URL dict gets the value
    'active': True. All other URL dicts get the value 'active': False.

    Note: this sets the 'active' values on the menu itself, not on a copy.
    """
    best_match = ''
    best_match_item = None

    for urlitem in menu:
        urlitem['active'] = False
        if not path.startswith(str(urlitem['url'])):
            continue

        match = os.path.commonprefix([str(urlitem['url']), path])
        if len(match) > len(best_match):
            best_match = match
            best_match_item = urlitem

    if best_match_item:
        best_match_item['active'] = True

    return menu


@register.inclusion_tag('subsubmenu.html', takes_context=True)
def show_subsubmenu(context, menu):
    """Mark the active menu item and display the subsubmenu"""
    menu = mark_active_menuitem(menu, context['request'].path)
    return {'subsubmenu': menu}


@register.filter(name='clients_of_type')
def clients_of_type(clients, client_type):
    """Filter and get clients of a particular type"""
    return clients_module.of_type(clients, client_type)


@register.filter(name='lookup')
def lookup(dictionary, key):
    """Get the value in the dictionary at given key"""
    return dictionary[key]


def _is_relative_url(url):
    """Check if the given link is relative or not."""
    parsed_url = urlparse(str(url))
    return not parsed_url.netloc


@register.filter(name='clients_get_platforms')
def clients_get_platforms(clients):
    """Return lists of self hosted platforms and all other platforms."""
    other = []
    web = []
    for client in clients:
        for platform in client['platforms']:
            if platform['type'] == 'web' and _is_relative_url(platform['url']):
                web.append(platform)
            else:
                other.append(platform)

    return {'web': web, 'other': other}



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

clients = [{
    'name': _('Roundcube'),
    'platforms': [{
        'type': 'web',
        'url': '/roundcube'
    }]
}]

backup = {
        'data': {
            'files': ['/etc/roundcube/freedombox-config.php',
                      '/var/lib/dbconfig-common/sqlite3/roundcube/roundcube']
        }
}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Generate DKIM keys for signing outgoing messages.

See: https://rspamd.com/doc/modules/dkim_signing.html
"""

import pathlib
import re
import shutil
import subprocess

from plinth.actions import privileged

_keys_dir = pathlib.Path('/var/lib/rspamd/dkim/')

DOMAIN_PART_REGEX = r'^[a-zA-Z0-9]([-a-zA-Z0-9]{,61}[a-zA-Z0-9])?$'


def _validate_domain_name(domain):
    for part in domain.split('.'):
        if not re.match(DOMAIN_PART_REGEX, part):
            raise ValueError('Invalid domain name')


@privileged
def get_dkim_public_key(domain: str) -> str:
    """Privileged action to get the public key from DKIM key."""
    _validate_domain_name(domain)
    key_file = _keys_dir / f'{domain}.dkim.key'
    output = subprocess.check_output(
        ['openssl', 'rsa', '-in',
         str(key_file), '-pubout'], stderr=subprocess.DEVNULL)
    return ''.join(output.decode().splitlines()[1:-1])


@privileged
def setup_dkim(domain: str):
    """Create DKIM key for a given domain."""
    _validate_domain_name(domain)

    _keys_dir.mkdir(exist_ok=True)
    _keys_dir.chmod(0o500)
    shutil.chown(_keys_dir, '_rspamd', '_rspamd')

    # Default path is /var/lib/dkim/$domain.$selector.key. Default selector is
    # "dkim". Use these to simplify key management until we have a need to
    # implement creating new or multiple keys.
    key_file = _keys_dir / f'{domain}.dkim.key'
    if key_file.exists():
        return

    # Ed25519 is widely *not* accepted as of 2022-01. See:
    # https://serverfault.com/questions/1023674
    subprocess.run([
        'rspamadm', 'dkim_keygen', '-t', 'rsa', '-b', '2048', '-s', 'dkim',
        '-d', domain, '-k', (str(key_file))
    ], check=True)
    key_file.chmod(0o400)



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

clients = [{
    'name': _('MediaWiki'),
    'platforms': [{
        'type': 'web',
        'url': '/mediawiki'
    }]
}]

backup = {
    'config': {
        'files': ['/etc/mediawiki/FreedomBoxSettings.php']
    },
    'data': {
        'directories': [
            '/var/lib/mediawiki-db/', '/var/lib/mediawiki/images/'
        ]
    },
    'services': ['mediawiki-jobrunner']
}



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox configuration form for OpenSSH server.
"""

from django import forms
from django.utils.translation import gettext_lazy as _


class SSHServerForm(forms.Form):
    """SSH server configuration form."""
    password_auth_disabled = forms.BooleanField(
        label=_('Disable password authentication'),
        help_text=_('Improves security by preventing password guessing. '
                    'Ensure that you have setup SSH keys in your '
                    'administrator user account before enabling this option.'),
        required=False,
    )

    allow_all = forms.BooleanField(
        label=_('Allow all users to login remotely'),
        help_text=_('Allow all users who have a valid account to login '
                    'remotely via SSH. When disabled, only users of groups '
                    'root, admin and freedombox-ssh can login via SSH.'),
        required=False,
    )



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure Shadowsocks Server."""

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.daemon import Daemon
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.package import Packages
from plinth.utils import format_lazy

from . import manifest, privileged

_description = [
    _('Shadowsocks is a tool for securely forwarding network requests to a'
      ' remote server. It consists of two parts: (1) a Shadowsocks server,'
      ' and (2) a Shadowsocks client with a SOCKS5 proxy.'),
    _('Shadowsocks can be used to bypass Internet filtering and '
      'censorship. This requires that the Shadowsocks server is in a '
      'location where it can freely access the Internet, without '
      'filtering.'),
    format_lazy(
        _('Your {box_name} can run a Shadowsocks server, that allows '
          'Shadowsocks clients to connect to it. Clients\' data will be '
          'encrypted and proxied through this server.'),
        box_name=_(cfg.box_name)),
]


class ShadowsocksServerApp(app_module.App):
    """FreedomBox app for Shadowsocks Server."""

    app_id = 'shadowsocksserver'

    _version = 1

    DAEMON = 'shadowsocks-libev-server@fbxserver'

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(
            app_id=self.app_id, version=self._version,
            name=_('Shadowsocks Server'), icon_filename='shadowsocks',
            short_description=_('Help Others Bypass Censorship'),
            description=_description, manual_page='Shadowsocks')
        self.add(info)

        menu_item = menu.Menu('menu-shadowsocks-server', info.name,
                              info.short_description, info.icon_filename,
                              'shadowsocksserver:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut(
            'shortcut-shadowsocks-server', info.name,
            short_description=info.short_description, icon=info.icon_filename,
            description=info.description, manual_page=info.manual_page,
            configure_url=reverse_lazy('shadowsocksserver:index'),
            login_required=True)
        self.add(shortcut)

        packages = Packages('packages-shadowsocks-server',
                            ['shadowsocks-libev'])
        self.add(packages)

        firewall = Firewall('firewall-shadowsocks-server', info.name,
                            ports=['shadowsocks-server-freedombox'],
                            is_external=True)
        self.add(firewall)

        daemon = Daemon(
            'daemon-shadowsocks-server', self.DAEMON,
            listen_ports=[(8388, 'tcp4'), (8388, 'tcp6'), (8388, 'udp4'),
                          (8388, 'udp6')])
        self.add(daemon)

        backup_restore = BackupRestore('backup-restore-shadowsocks-server',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if not old_version:
            self.enable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox privacy app."""

from django import forms
from django.utils.translation import gettext_lazy as _

from plinth import cfg
from plinth.utils import format_lazy


class PrivacyForm(forms.Form):
    """Privacy configuration form."""

    enable_popcon = forms.BooleanField(
        label=_('Periodically submit a list of apps used (suggested)'),
        required=False, help_text=format_lazy(
            _('Help Debian/{box_name} developers by participating in the '
              'Popularity Contest package survey program. When enabled, a '
              'list of apps used on this system will be anonymously submitted '
              'to Debian every week. Statistics for the data collected are '
              'publicly available at <a href="https://popcon.debian.org/" '
              'target="_blank">popcon.debian.org</a>. Submission happens over '
              'the Tor network for additional anonymity if Tor app is enabled.'
              ), box_name=_(cfg.box_name)))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Functional, browser based tests for privacy app."""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.system, pytest.mark.privacy]


class TestPrivacyApp(functional.BaseAppTests):
    """Tests for privacy app."""

    app_name = 'privacy'
    has_service = False
    has_web = False
    disable_after_tests = False

    @pytest.fixture(autouse=True)
    def fixture_background(self, session_browser):
        """Login, install, and enable the app."""
        functional.login(session_browser)
        functional.nav_to_module(session_browser, self.app_name)
        yield

    def test_enable_disable(self, session_browser):
        """Skip test for enabling and disabling the app."""
        pytest.skip('Can not be disabled')

    @pytest.mark.backups
    def test_enable_disable_popcon(self, session_browser):
        """Test that popcon can be enable/disabled."""
        functional.change_checkbox_status(session_browser, self.app_name,
                                          'id_enable_popcon', 'disabled')
        functional.change_checkbox_status(session_browser, self.app_name,
                                          'id_enable_popcon', 'enabled')
        assert session_browser.find_by_id('id_enable_popcon').checked
        functional.change_checkbox_status(session_browser, self.app_name,
                                          'id_enable_popcon', 'disabled')
        assert not session_browser.find_by_id('id_enable_popcon').checked

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test that backup and restore operations work on the app."""
        functional.change_checkbox_status(session_browser, self.app_name,
                                          'id_enable_popcon', 'disabled')
        functional.backup_create(session_browser, self.app_name,
                                 'test_' + self.app_name)
        functional.nav_to_module(session_browser, self.app_name)
        functional.change_checkbox_status(session_browser, self.app_name,
                                          'id_enable_popcon', 'enabled')
        functional.backup_restore(session_browser, self.app_name,
                                  'test_' + self.app_name)
        functional.nav_to_module(session_browser, self.app_name)
        assert not session_browser.find_by_id('id_enable_popcon').checked

    def test_uninstall(self, session_browser):
        """Skip test for uninstall."""
        pytest.skip('Essential app')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for snapshot module."""

import urllib.parse

from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy

from plinth import app as app_module
from plinth.modules import snapshot as snapshot_module
from plinth.modules import storage
from plinth.views import AppView, messages_error

from . import get_configuration, privileged
from .forms import SnapshotForm

# i18n for snapshot descriptions
SNAPSHOT_DESCRIPTION_STRINGS = {
    'manually created': gettext_lazy('manually created'),
    'timeline': gettext_lazy('timeline'),
    'apt': gettext_lazy('apt'),
}

subsubmenu = [
    {
        'url': reverse_lazy('snapshot:index'),
        'text': gettext_lazy('Configure')
    },
    {
        'url': reverse_lazy('snapshot:manage'),
        'text': gettext_lazy('Manage Snapshots')
    },
]


def not_supported_view(request):
    """Show that snapshots are not supported on the system."""
    app = app_module.App.get('snapshot')
    template_data = {
        'app_info': app.info,
        'title': app.info.name,
        'fs_type': storage.get_filesystem_type(),
        'fs_types_supported': snapshot_module.fs_types_supported,
    }
    return TemplateResponse(request, 'snapshot_not_supported.html',
                            template_data)


class SnapshotAppView(AppView):
    """Show snapshot app main page."""

    app_id = 'snapshot'
    template_name = 'snapshot.html'
    form_class = SnapshotForm

    def dispatch(self, request, *args, **kwargs):
        """If snapshots are not supported, use a different view."""
        if not snapshot_module.is_supported():
            return not_supported_view(request)

        return super().dispatch(request, *args, **kwargs)

    def get_initial(self):
        """Return the values to fill in the form."""
        initial = super().get_initial()
        initial.update(get_configuration())
        return initial

    def get_context_data(self, *args, **kwargs):
        """Add additional context data for template."""
        context = super().get_context_data(*args, **kwargs)
        context['subsubmenu'] = subsubmenu
        return context

    def form_valid(self, form):
        """Apply the changes submitted in the form."""
        if 'update' in self.request.POST:
            update_configuration(self.request, form.initial, form.cleaned_data)

        return super().form_valid(form)


def manage(request):
    """Show snapshot list."""
    if not snapshot_module.is_supported():
        return not_supported_view(request)

    if request.method == 'POST':
        if 'create' in request.POST:
            privileged.create()
            messages.success(request, _('Created snapshot.'))
        if 'delete_selected' in request.POST:
            to_delete = request.POST.getlist('snapshot_list')
            if to_delete:
                # Send values using GET params instead of session variables so
                # that snapshots can be deleted even when disk is full.
                params = [('snapshots', number) for number in to_delete]
                params = urllib.parse.urlencode(params)
                url = reverse('snapshot:delete-selected')
                return HttpResponseRedirect(f'{url}?{params}')

    snapshots = privileged.list_()
    has_deletable_snapshots = any([
        snapshot for snapshot in snapshots
        if not snapshot['is_default'] and not snapshot['is_active']
    ])

    app = app_module.App.get('snapshot')
    return TemplateResponse(
        request, 'snapshot_manage.html', {
            'title': app.info.name,
            'app_info': app.info,
            'snapshots': snapshots,
            'has_deletable_snapshots': has_deletable_snapshots,
            'subsubmenu': subsubmenu,
        })


def update_configuration(request, old_status, new_status):
    """Update configuration of snapshots."""

    def make_config(args):
        key, stamp = args[0], args[1]
        if old_status[key] != new_status[key]:
            if 'limit' in key:
                return stamp.format('0-{}'.format(new_status[key]))

            return stamp.format(new_status[key])

        return None

    config = filter(
        None,
        map(make_config, [
            ('enable_timeline_snapshots', 'TIMELINE_CREATE={}'),
            ('hourly_limit', 'TIMELINE_LIMIT_HOURLY={}'),
            ('daily_limit', 'TIMELINE_LIMIT_DAILY={}'),
            ('weekly_limit', 'TIMELINE_LIMIT_WEEKLY={}'),
            ('monthly_limit', 'TIMELINE_LIMIT_MONTHLY={}'),
            ('yearly_limit', 'TIMELINE_LIMIT_YEARLY={}'),
            ('free_space', 'FREE_LIMIT={}'),
        ]))

    if old_status['enable_software_snapshots'] != new_status[
            'enable_software_snapshots']:
        if new_status['enable_software_snapshots'] == 'yes':
            privileged.disable_apt_snapshot('no')
        else:
            privileged.disable_apt_snapshot('yes')

    try:
        privileged.set_config(list(config))
        messages.success(request, _('Configuration updated.'))
    except Exception as exception:
        messages_error(request, _('Configuration update failed.'), exception)


def delete_selected(request):
    """View to delete selected snapshots."""
    if request.method == 'POST':
        to_delete = set(request.POST.getlist('snapshots'))
    else:
        to_delete = set(request.GET.getlist('snapshots'))

    if not to_delete:
        return redirect(reverse('snapshot:manage'))

    snapshots = privileged.list_()
    snapshots_to_delete = [
        snapshot for snapshot in snapshots if snapshot['number'] in to_delete
        and not snapshot['is_active'] and not snapshot['is_default']
    ]

    if request.method == 'POST':
        try:
            for snapshot in snapshots_to_delete:
                privileged.delete(snapshot['number'])

            messages.success(request, _('Deleted selected snapshots'))
        except Exception as exception:
            message = _('Deleting snapshot failed.')
            stderr = getattr(exception, 'stderr', b'').decode()
            if 'Config is in use.' in stderr:
                message = _('Snapshot is currently in use. '
                            'Please try again later.')

            messages_error(request, message, exception)

        return redirect(reverse('snapshot:manage'))

    return TemplateResponse(request, 'snapshot_delete_selected.html', {
        'title': _('Delete Snapshots'),
        'snapshots': snapshots_to_delete
    })


def rollback(request, number):
    """Show confirmation to rollback to a snapshot."""
    if request.method == 'POST':
        privileged.rollback(number)
        messages.success(
            request,
            _('Rolled back to snapshot #{number}.').format(number=number))
        messages.warning(
            request,
            _('The system must be restarted to complete the rollback.'))
        return redirect(reverse('power:restart'))

    snapshots = privileged.list_()
    snapshot = None
    for current_snapshot in snapshots:
        if current_snapshot['number'] == number:
            snapshot = current_snapshot

    return TemplateResponse(request, 'snapshot_rollback.html', {
        'title': _('Rollback to Snapshot'),
        'snapshot': snapshot
    })



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Unit tests for helpers of I2P application.
"""

from pathlib import Path

import pytest

from plinth.modules.i2p.helpers import TunnelEditor

DATA_DIR = Path(__file__).parent / 'data'
TUNNEL_CONF_PATH = DATA_DIR / 'i2ptunnel.config'
TUNNEL_HTTP_NAME = 'I2P HTTP Proxy'


@pytest.fixture(name='editor')
def fixture_editor():
    """Setup editor for each test."""
    return TunnelEditor(str(TUNNEL_CONF_PATH))


def test_reading_conf(editor):
    """Test reading configuration file."""
    editor.read_conf()
    assert len(editor.lines) > 1


def test_setting_idx(editor):
    """Test setting index for editing a tunnel."""
    editor.read_conf()
    assert editor.idx is None
    editor.set_tunnel_idx(TUNNEL_HTTP_NAME)
    assert editor.idx == 0


def test_setting_tunnel_props(editor):
    """Test setting a tunnel property."""
    editor.read_conf()
    editor.set_tunnel_idx('I2P HTTP Proxy')
    interface = '0.0.0.0'
    editor.set_tunnel_prop('interface', interface)
    assert editor['interface'] == interface


def test_getting_nonexistent_props(editor):
    """Test getting nonexistent property."""
    editor.read_conf()
    editor.idx = 0
    with pytest.raises(KeyError):
        _ = editor['blabla']  # noqa


def test_setting_new_props(editor):
    """Test setting new properties."""
    editor.read_conf()
    editor.idx = 0
    value = 'lol'
    prop = 'blablabla'
    editor[prop] = value
    assert editor[prop] == value



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Shutdown/restart the system."""

import subprocess

from plinth.actions import privileged


@privileged
def restart():
    """Restart the system."""
    subprocess.call('reboot')


@privileged
def shutdown():
    """Shut down the system."""
    subprocess.call(['shutdown', 'now'])



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Forms for the Matrix Synapse module.
"""

from django import forms
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth.modules.coturn.forms import turn_uris_validator
from plinth.utils import format_lazy

_registration_verification_choices = [
    ('disabled',
     _('Disabled. This could lead to adversaries registering many spam '
       'accounts on your server with automated scripts.')),
    ('token',
     _('Require users creating a new account to use a registration token. A '
       'token will be created automatically. Pass this token to your '
       'potential new users. They will be asked for the token during '
       'registration. (recommended)')),
]


class MatrixSynapseForm(forms.Form):
    enable_public_registration = forms.BooleanField(
        label=_('Enable Public Registration'), required=False, help_text=_(
            'Enabling public registration means that anyone on the Internet '
            'can register a new account on your Matrix server. Disable this '
            'if you only want existing users to be able to use it.'))

    registration_verification = forms.ChoiceField(
        label=_('Verification method for registration'),
        choices=_registration_verification_choices, required=True,
        widget=forms.RadioSelect)

    enable_managed_turn = forms.BooleanField(
        label=_('Automatically manage audio/video call setup'), required=False,
        help_text=format_lazy(
            _('Configures the local <a href={coturn_url}>coturn</a> app as '
              'the STUN/TURN server for Matrix Synapse. Disable this if you '
              'want to use a different STUN/TURN server.'),
            coturn_url=reverse_lazy('coturn:index')))

    # STUN/TURN server setup
    turn_uris = forms.CharField(
        label=_('STUN/TURN Server URIs'), required=False, strip=True,
        widget=forms.Textarea(attrs={'rows': 4}),
        help_text=_('List of public URIs of the STUN/TURN server, one on each '
                    'line.'), validators=[turn_uris_validator])

    shared_secret = forms.CharField(
        label=_('Shared Authentication Secret'), required=False, strip=True,
        help_text=_('Shared secret used to compute passwords for the '
                    'TURN server.'))

    def clean_turn_uris(self):
        """Normalize newlines in URIs."""
        data = self.cleaned_data['turn_uris']
        return '\n'.join([uri.strip() for uri in data.splitlines()])



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for sso app.
"""

import pytest
from plinth.tests import functional

pytestmark = [pytest.mark.system, pytest.mark.essential, pytest.mark.sso]


@pytest.fixture(scope='module', autouse=True)
def fixture_background(session_browser):
    """Login and install the app."""
    functional.login(session_browser)
    functional.install(session_browser, 'syncthing')
    functional.app_enable(session_browser, 'syncthing')
    yield
    functional.app_disable(session_browser, 'syncthing')


def test_app_access(session_browser):
    """Test that only logged-in users can access Syncthing web interface."""
    functional.logout(session_browser)
    functional.access_url(session_browser, 'syncthing')
    assert functional.is_login_prompt(session_browser)

    functional.login(session_browser)
    functional.access_url(session_browser, 'syncthing')
    assert functional.is_available(session_browser, 'syncthing')



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.contrib import messages
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth.forms import TLSDomainForm
from plinth.modules import quassel
from plinth.views import AppView


class QuasselAppView(AppView):
    app_id = 'quassel'
    form_class = TLSDomainForm

    def get_initial(self):
        """Return the values to fill in the form."""
        initial = super().get_initial()
        initial['domain'] = quassel.get_domain()
        return initial

    def form_valid(self, form):
        """Change the access control of Radicale service."""
        data = form.cleaned_data
        if quassel.get_domain() != data['domain']:
            quassel.set_domain(data['domain'])
            app = app_module.App.get('quassel')
            app.get_component('letsencrypt-quassel').setup_certificates()
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the Help module
"""

from django.urls import re_path

from plinth.utils import non_admin_view

from . import views

urlpatterns = [
    re_path(r'^help/$', non_admin_view(views.index), name='index'),
    re_path(r'^help/about/$', non_admin_view(views.about), name='about'),
    re_path(r'^help/feedback/$', non_admin_view(views.feedback),
            name='feedback'),
    re_path(r'^help/support/$', non_admin_view(views.support), name='support'),
    re_path(r'^help/contribute/$', non_admin_view(views.contribute),
            name='contribute'),
    re_path(r'^help/manual/$', non_admin_view(views.manual), name='manual'),
    re_path(r'^help/manual/(?P<lang>\w*(-\w*)?)/$',
            non_admin_view(views.manual), name='manual'),
    re_path(r'^help/manual/(?P<lang>\w*(-\w*)?)/(?P<page>[\w-]+)$',
            non_admin_view(views.manual), name='manual-page'),
    re_path(r'^help/manual-download/$', non_admin_view(views.download_manual),
            name='download-manual'),
    re_path(r'^help/status-log/$', non_admin_view(views.status_log),
            name='status-log'),
]
